// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#ifndef Ausm2D_HH
#define Ausm2D_HH

class Ausm2D;

#include <CompressibleFFlux2D.hh>

/** */
class Ausm2D : public CompressibleFFlux2D
{
public:
  /** */ Ausm2D(carg_t arg);
  /** */ void operate(size_t ie, size_t ip, size_t from, size_t to, 
		      real nx, real ny, bool mod_from, bool mod_to);
};

inline void Ausm2D::operate(size_t ie, size_t ip, 
			    size_t from, size_t to, 
			    real nx, real ny, 
			    bool mod_from, bool mod_to)
{
  //.. compute normed interface-normal vector
  real area = sqrt(nx*nx + ny*ny);
  real i_area = 1./area;
  real hnx = nx * i_area;
  real hny = ny * i_area;
  
  //.. compute normal speed and speed of sound (all central)
  real i_rho_h = 1./rho_av[ie];
  real uuh=(rhou_av[ie] * hnx + rhov_av[ie] * hny) * i_rho_h;
  real q_rho_uv = rhou_av[ie] * rhou_av[ie] + rhov_av[ie] * rhov_av[ie];
  real a_speed = ((rhoE_av[ie] * rho_av[ie] -0.5 * q_rho_uv)*gamma2) * (i_rho_h * i_rho_h);
  if(a_speed < 0.1)
    a_speed=0.1;    // avoid division by 0 if critical
  a_speed = sqrt(a_speed);
  
  //.. compute limited Characteristics
  real ch_to = uuh + a_speed;
  if(ch_to < 0.0)
    ch_to = 0.0;
  real ch_from = uuh - a_speed;
  if(ch_from > 0.0)
    ch_from = 0.0;
  //.. compute characteristic coefficients
  real ch_diff = ch_to - ch_from;
  real i_ch_diff = 1./ch_diff;
  real ch_to_norm  = ch_to * i_ch_diff;
  real ch_from_norm = ch_from * i_ch_diff;
  
  //.. compute pressure at _to side
  real p_to_h = gamma1 
    * (rhoE_to[ie] - 0.5 *(rhou_to[ie] * rhou_to[ie] + rhov_to[ie] * rhov_to[ie]) / rho_to[ie]);
  
  //.. compute pressure at _from side
  real p_from_h = gamma1 
    * (rhoE_from[ie] - 0.5 *(rhou_from[ie] * rhou_from[ie] + rhov_from[ie] * rhov_from[ie])
       / rho_from[ie]);
  
  //.. compute combined pressure
  real ph = ch_to_norm * p_from_h - ch_from_norm * p_to_h;
  //.. treat convection speed as the pressure (recompute it)
  
  //.. decide for convective fluxes and compute total ausm flux
  real uuh_area = uuh * area;
  real flux_rho, flux_rhou, flux_rhov, flux_rhoE;
  if(uuh >= 0.0) {
    flux_rho = rho_from[ie] * uuh_area;
    flux_rhou = rhou_from[ie] * uuh_area + ph*nx;
    flux_rhov = rhov_from[ie] * uuh_area + ph*ny;
    flux_rhoE = (rhoE_from[ie] + p_from_h) * uuh_area;
  } else {
    flux_rho = rho_to[ie] * uuh_area;
    flux_rhou = rhou_to[ie] * uuh_area + ph*nx;
    flux_rhov = rhov_to[ie] * uuh_area + ph*ny;
    flux_rhoE = (rhoE_to[ie] + p_to_h) * uuh_area;
  };
  
  //.. Add to pointwise residual
  if (mod_to) {
    res_rho[to] += flux_rho;
    res_rhou[to] += flux_rhou;
    res_rhov[to] += flux_rhov;
    res_rhoE[to] += flux_rhoE;
  };
  if (mod_from) {
    res_rho[from] -= flux_rho;
    res_rhou[from] -= flux_rhou;
    res_rhov[from] -= flux_rhov;
    res_rhoE[from] -= flux_rhoE;
  };
};

#endif










