// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik
// Universitaet Duisburg, Germany
// Institute for Combustion and Gas Dynamics
// University of Duisburg, Germany
// Sun May 16 1999
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!
#ifndef CompressibleBSplitFlux2D_HH
#define CompressibleBSplitFlux2D_HH

#include <mouse_config.h>

class CompressibleBSplitFlux2D;

#include <mouse_types.hh>
#include <ConstructArg.hh>
#include <CompressibleOperator2D.hh>
#include <Roe2D.hh>

/**
*/
class CompressibleBSplitFlux2D : public CompressibleOperator2D
{
protected:
  /** */ void compute_values(size_t i, real nx, real ny, 
			     real &rho_out, real &rhou_out, real &rhov_out, real &rhoE_out);
  /** */ void compute_flux(size_t node, real nx, real ny);

public:
  /** */ CompressibleBSplitFlux2D (carg_t arg);
  /** */ virtual ~CompressibleBSplitFlux2D ();
  /** */ void operate(size_t ie, size_t ip, size_t from, size_t to, 
		      real nx, real ny, bool mod_from, bool mod_to);
};

inline void CompressibleBSplitFlux2D::compute_values(size_t i, real nx, real ny, 
						     real &rho_out, real &rhou_out, 
						     real &rhov_out, real &rhoE_out)
{
  //.. Compute normal Mach number to decide
  real uv_n = (rhou[i] * nx + rhov[i] * ny) / rho[i];
  real q_roh_uv = rhou[i] * rhou[i] + rhov[i] * rhov[i];
  real a_speed = ((rhoE[i] * rho[i] -0.5 * q_roh_uv) *gamma2) / (rho[i] * rho[i]);

  // avoid division by 0 if critical
  if (a_speed < 0.1) a_speed=0.1;

  a_speed = sqrt(a_speed);
  real mach = uv_n / a_speed;

  //.. Decide upon Mach number
  //.... supersonic outflow, do nothing
  if (mach <= -1) {
    rho_out = rho[i];
    rhou_out = rhou[i];
    rhov_out = rhov[i];
    rhoE_out = rhoE[i];
  }
  //.... supersonic inflow, take everything
  else if (mach >= 1) {
    rho_out = rho_oo;
    rhou_out = rhou_oo;
    rhov_out = rhov_oo;
    rhoE_out = rhoE_oo;
  }
  //.... subsonic outflow, use pressure from outside to adjust energy
  else if (mach < 0) {
    rho_out = rho[i];
    rhou_out = rhou[i];
    rhov_out = rhov[i];
    rhoE_out = p_oo * igamma1 + 0.5 * q_roh_uv / rho[i];
  }
  //.... subsonic inflow, take all from outside apart pressure
  //     stability problems -> keep Energy from inside
  else if (mach >= 0) {
    rho_out = rho_oo;
    rhou_out = rhou_oo;
    rhov_out = rhov_oo;
    rhoE_out = rhoE[i];
  }
  //.... Error condition for corrupted mach number
  else {
    Error("corrupted Mach number");
  };
};

inline void CompressibleBSplitFlux2D::compute_flux(size_t node, real nx, real ny)
{
  real flux_rho, flux_rhou, flux_rhov, flux_rhoE;
  real rho_cell = rho[node];
  real rhou_cell = rhou[node];
  real rhov_cell = rhov[node];
  real rhoE_cell = rhoE[node];
  //real rho_out, rhou_out, rhov_out, rhoE_out;
  //compute_values(node, nx, ny, rho_out, rhou_out, rhov_out, rhoE_out);
  Roe2D::compute(rho_oo, rhou_oo, rhov_oo, rhoE_oo,
		 rho_cell, rhou_cell, rhov_cell, rhoE_cell,
		 nx, ny, gamma,
		 flux_rho, flux_rhou, flux_rhov, flux_rhoE);
  res_rho[node] += flux_rho;
  res_rhou[node] += flux_rhou;
  res_rhov[node] += flux_rhov;
  res_rhoE[node] += flux_rhoE;
};

inline void CompressibleBSplitFlux2D::operate(size_t ie, size_t ip, size_t from, size_t to, 
					      real nx, real ny, bool mod_from, bool mod_to)
{
  real flux_rho, flux_rhou, flux_rhov, flux_rhoE;
  if (mod_from) compute_flux(from, nx, ny);
  if (mod_to) compute_flux(to, nx, ny);
};


#endif
