// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik
// Universitaet Duisburg, Germany
// Institute for Combustion and Gas Dynamics
// University of Duisburg, Germany
// Mon Jun 7 1999
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include "LeastSquareInterpolator2D.hh"

LeastSquareInterpolator2D::LeastSquareInterpolator2D()
{
  F = new TList<real>(5,5);
  W = new TList<real>(F);
  X_list = new TList<real>(F);
  Y_list = new TList<real>(F);
  M_uptodate = false;
  RHS_uptodate = false;
};

LeastSquareInterpolator2D::~LeastSquareInterpolator2D()
{
  delete X_list;
  delete Y_list;
  delete F;
  delete W;
};

void LeastSquareInterpolator2D::ComputeMatrix()
{
  for (int i = 0; i < num_coeff; i++) {
    for (int j = 0; j < num_coeff; j++) {
      M[i][j] = 0;
    };
  };
  FORALL(i,W->) {
    real x = X(i);
    real y = Y(i);
    real w = W->At(i);
    /*
    M[0][0] += 2.0*w;      
    M[0][1] += 2.0*w*x;
    M[0][2] += 2.0*w*y;
    M[1][0] += 2.0*w*x;
    M[1][1] += 2.0*w*x*x;
    M[1][2] += 2.0*w*y*x;
    M[2][0] += 2.0*w*y;
    M[2][1] += 2.0*w*y*x;
    M[2][2] += 2.0*w*y*y;
    */

    M[0][0] += 2.0*w;
    M[0][1] += 2.0*w*x;
    M[0][2] += 2.0*w*y;
    M[0][3] += 2.0*w*x*x;
    M[0][4] += 2.0*w*y*y;
    M[1][0] += 2.0*w*x;
    M[1][1] += 2.0*w*x*x;
    M[1][2] += 2.0*w*y*x;
    M[1][3] += 2.0*w*x*x*x;
    M[1][4] += 2.0*w*y*y*x;
    M[2][0] += 2.0*w*y;
    M[2][1] += 2.0*w*y*x;
    M[2][2] += 2.0*w*y*y;
    M[2][3] += 2.0*w*x*x*y;
    M[2][4] += 2.0*w*y*y*y;
    M[3][0] += 2.0*w*x*x;
    M[3][1] += 2.0*w*x*x*x;
    M[3][2] += 2.0*w*x*x*y;
    M[3][3] += 2.0*w*x*x*x*x;
    M[3][4] += 2.0*w*y*y*x*x;
    M[4][0] += 2.0*w*y*y;
    M[4][1] += 2.0*w*y*y*x;
    M[4][2] += 2.0*w*y*y*y;
    M[4][3] += 2.0*w*y*y*x*x;
    M[4][4] += 2.0*w*y*y*y*y;
  };
  M_inv = M.Inv();
};

void LeastSquareInterpolator2D::ComputeRhs()
{
  for (int i = 0; i < num_coeff; i++) {
    RHS[i] = 0.0;
  };
  FORALL(i,W->) {
    real x = X(i);
    real y = Y(i);
    real w = W->At(i);
    real f = F->At(i); 
    /*
    RHS[0] += 2.0*w*f;
    RHS[1] += 2.0*w*x*f;
    RHS[2] += 2.0*w*y*f;
    */
    
    RHS[0] += 2.0*w*f;
    RHS[1] += 2.0*w*x*f;
    RHS[2] += 2.0*w*y*f;
    RHS[3] += 2.0*w*x*x*f;
    RHS[4] += 2.0*w*y*y*f;
  };
};

void LeastSquareInterpolator2D::ComputeCoeff()
{
  for (int i = 0; i < num_coeff; i++) {
    COEFF[i] = 0;
    for (int j = 0; j < num_coeff; j++) {
      COEFF[i] += M_inv[i][j]*RHS[j];
    };
  };
};

void LeastSquareInterpolator2D::Reset()
{
  F->DelAll();
  M_uptodate = false;
};

size_t LeastSquareInterpolator2D::AddPoint(real x, real y)
{
  size_t i = F->AddEntry();
  X_list->At(i) = x;
  Y_list->At(i) = y;
  F->At(i) = 0;
  W->At(i) = 1;
  M_uptodate = false;
  return i;
};

void LeastSquareInterpolator2D::SetValue(size_t i, real f)
{
  F->At(i) = f;
  RHS_uptodate = false;
};

void LeastSquareInterpolator2D::SetWeight(size_t i, real w)
{
  W->At(i) = w;
  M_uptodate = false;
};

void LeastSquareInterpolator2D::SetPosition(real x, real y)
{
  x0 = x;
  y0 = y;
  FORALL(i,W->) {
    W->At(i) = 1/((x-X(i))*(x-X(i))+(y-Y(i))*(y-Y(i)));
    W->At(i) = 1;
  };
  M_uptodate = false;
};

real LeastSquareInterpolator2D::ComputeValue(real x, real y)
{
  if (!M_uptodate) {
    ComputeMatrix();
    M_uptodate = true;
    RHS_uptodate = false;
  };
  if (!RHS_uptodate) {
    ComputeRhs();
    ComputeCoeff();
    RHS_uptodate = true;
  };
  //cout << M << endl;
  //cout << M_inv << endl;
//   real value = 0;
  for (uint i = 0; i < num_coeff; i++) cout << "C[" << i << "]=" << COEFF[i] << endl;
  return COEFF[0] + COEFF[1]*x + COEFF[2]*y + COEFF[3]*x*x + COEFF[4]*y*y;
};
