// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include "List.hh"
#include <iostream.h>
#include <stdlib.h>

//
//.. constructor
//
void List::InitList (size_t a_max_num_entries, size_t a_delta_entries)
{
  if (initialized) {
    cout << "fatal error: attempt to initialize a List twice" << endl;
    exit (EXIT_FAILURE);
  };
  num_entries = 0;
  max_num_entries = a_max_num_entries;
  delta_entries = a_delta_entries;
  offset = NULL;
  if (master == this) {
    //active = new BitArray(max_num_entries);
    active = new bool[max_num_entries];
    for (size_t i_entry = 0; i_entry < max_num_entries; i_entry++) {
      //active->SetBit(i_entry, false);
      active[i_entry] = false;
    };
  } else {
    active = NULL;
  };
  client = NULL;
  num_clients = 0;
  initialized = true;
};

List::List ()
{
  client = NULL;
  master = this;
  initialized = false;
  num_clients = 0;
};

List::List (size_t a_max_num_entries, size_t a_delta_entries)
{
  is_clean = true;
  initialized = false;
  master = this;
  InitList (a_max_num_entries, a_delta_entries);
  num_clients = 0;
};

List::List (List *linked_list) 
{
  client = NULL;
  initialized = false;
  master = linked_list->master;
  InitList (master->MaxNumEntries(), master->DeltaEntries());
  master->AddClient(this);
  num_clients = 0;
};

//
//.. Link
//
void List::Link (List *linked_list) 
{
  if (this != master) {
    cout << "This 'List' is already linked to another 'List'!" << endl;
    exit(EXIT_FAILURE);
  };
  if (num_clients != 0) {
    cout << "This 'List' is already a master for someone!" << endl;
    exit(EXIT_FAILURE);
  };
  client = NULL;
  //initialized = false;
  master = linked_list->master;
  DelAll();
  InitList (master->MaxNumEntries(), master->DeltaEntries());
  //while (NumEntries() < master->NumEntries()) {
  //size_t i = AddEntry();
  FORALL(i, this->) {
    InitEntry(i);
  };
  master->AddClient(this);
};

//
//.. AddClient
//
void List::AddClient (List *client_to_add) 
{
  if (this != master) {
    cout << "fatal error trying to link to a non master list!" << endl;
    exit(EXIT_FAILURE);
  };
  List **new_client;
  new_client = new List* [num_clients + 1];
  for (unsigned int i = 0; i < num_clients; i++) new_client[i] = client[i];
  new_client[num_clients] = client_to_add;
  num_clients++;
  delete [] client;
  client = new_client;
  while (client_to_add->num_entries < num_entries) {
    client_to_add->InitEntry(client_to_add->num_entries);
    client_to_add->num_entries++;
  };
};

//
//.. DelClient
//
void List::DelClient (List *client_to_del) {
  if (this != master) {
    cout << "fatal error trying to delete a client from a non master list!" << endl;
    exit(EXIT_FAILURE);
  };
  List **new_client;
  int off = 0;
  for (unsigned int i = 0; i < num_clients - off; i++) {
    if (client[i] == client_to_del) off += 1;
    client[i] = client[i + off];
  };
  new_client = new List* [num_clients - off];
  for (unsigned int i = 0; i < num_clients- off; i++) new_client[i] = client[i];
  num_clients -= off;
  delete [] client;
  client = new_client;
};

//
//.. destructor
//
List::~List () 
{
  if (this != master) {
    master->DelClient(this);
  } else {
    for (size_t i = 0; i < num_clients; i++) {
      client[i]->master = client[i];
    };
  };
  delete [] offset;
  delete [] client;
  delete [] active;
};

//
//.. NumActiveEntries
//
size_t List::NumActiveEntries () const
{
  size_t num_actives = 0;
  for (size_t i_entry = begin(); i_entry < num_entries; i_entry = next(i_entry)) {
    num_actives++;
  };
  return num_actives;
};

//
//.. ExtendList
//
void List::ExtendList (size_t delta) 
{
  if (this != master) {
    cout << "fatal error in ExtendList this != master." << endl;
    exit(EXIT_FAILURE);
  };
  bool *new_active;
  Extend (delta);
  max_num_entries += delta;
  new_active = new bool[max_num_entries];
  for (size_t i = 0; i < max_num_entries; i++) {
    if (i < num_entries) {
      new_active[i] = active[i];
    } else {
      new_active[i] = false;
    }
  };
  delete [] active;
  active = new_active;
  for (unsigned int j = 0; j < num_clients; j++) {
    client[j]->Extend(delta);
    client[j]->max_num_entries += delta;
  };
};

//
//.. AddEntry
//
size_t List::AddEntry () {
  size_t new_entry;
  if (this != master) {
    // this is not the master
    new_entry = master->AddEntry();
  } else { 
    // this is the master !
    if (num_entries == max_num_entries) master->ExtendList(delta_entries);
    //active->SetBit(num_entries, 1);
    active[num_entries] = true;
    num_entries++;
    InitEntry(num_entries - 1);
    for (unsigned int i = 0; i < num_clients; i++) {
      client[i]->num_entries++;
      if (client[i]->num_entries != num_entries) {
	cout << "fatal List syncronization error!" << endl;
	exit(EXIT_FAILURE);
      };
      client[i]->InitEntry(num_entries - 1);
    };
  };
  return num_entries - 1;
};

//
//.. DelEntry
//
void List::DelEntry (size_t i_entry) {
  /*
  if (master == this) active->SetBit(i_entry, 0);
  else master->DelEntry(i_entry);
  */
  //master->active->SetBit(i_entry, 0);
  master->active[i_entry] = false;
  is_clean = false;
};

//
//.. DelAll
//
void List::DelAll () {
  if (initialized) {
    for (size_t i_entry = 0; i_entry < max_num_entries; i_entry++) DelEntry (i_entry);
    CleanUp();
  };
};

size_t List::begin() const
{
  size_t i = 0;
  while ((i < num_entries) && !IsActive(i)) i++;
  return i;
};

//
//.. CreateOffsetList
//
void List::CreateOffsetList () {
  if (master == this) {
    max_offset = 0;
    DeleteOffsetList();
    offset = new size_t [num_entries];
    for (size_t i_entry = 0; i_entry < num_entries; i_entry++) {
      offset[i_entry] = max_offset;
      if (!IsActive(i_entry)) max_offset++;
    };
  } else {
    master->CreateOffsetList();
  };
};

//
//.. Offset
//
size_t List::Offset (size_t i_entry) {
  if (master->offset == NULL) {
    cerr << "error: no offset information available" << endl;
    cerr << "       at this point in the program\n" << endl;
    exit (EXIT_FAILURE);
  } else {
    return master->offset[i_entry];
  };
  return 0; // dummy
};

//
//.. CopyEntry
//
void List::CopyEntry(size_t src, size_t dest)
{
  if (master == this) {
    for (unsigned int i_client = 0; i_client < num_clients; i_client++) {
      client[i_client]->CopyEntry(src, dest);
    };
  };
};

//
//.. CleanUp
//
void List::CleanUp () {
  //cout << "cleaning" << endl;
  //exit(EXIT_FAILURE);
  size_t i_entry;
  if (master == this) {
    CreateOffsetList();
    for (i_entry = 0; i_entry < num_entries; i_entry++) {
      if (Offset(i_entry) != 0) {
	CopyEntry (i_entry, i_entry - Offset(i_entry));
	/*
	for (i_client = 0; i_client < num_clients; i_client++) {
	  client[i_client]->CopyEntry(i_entry, i_entry - Offset(i_entry));
	};
	*/
      };
      if (i_entry < num_entries - max_offset) {
	//active->SetBit(i_entry, 1);
	active[i_entry] = true;
      } else {
	//active->SetBit(i_entry, 0);
	active[i_entry] = false;
      };
    };
    if (num_entries != 0) num_entries -= max_offset;
    for (unsigned int i_client = 0; i_client < num_clients; i_client++) {
      client[i_client]->num_entries = num_entries;
    };
    is_clean = true;
  } else {
    master->CleanUp();
  };
};

//
//.. operator=
//
void List::operator=(const List &other)
{
  delete [] offset;
  delete [] active;
  num_entries = other.num_entries;
  max_num_entries = other.max_num_entries;
  master->AddClient(this);
  if (this != master) {
    master->DelClient(this);
    master = other.master;
    delta_entries = other.delta_entries;
  };
  //active = new BitArray(max_num_entries);
  active = new bool [max_num_entries];
  for(size_t i = 0; i < max_num_entries; i++) {
    //active->SetBit(i, other.active->GetBit(i));
    active[i] = other.active[i];
  };
};

//
//.. SetDelta
//
void List::SetDelta(size_t new_delta)
{
  if (this != master) {
    cerr << "cannot modify the delta value for client lists" << endl;
    exit(EXIT_FAILURE);
  };
  delta_entries = new_delta;
};







