// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include "MObject.hh"
#include <Communicator.hh>
#include <Parser.hh>
#include <ConstructArg.hh>
#include <BroadcastEvent.hh>
#include <RequestEvent.hh>
#include <MessageEvent.hh>
#include <MStringList.hh>
#include <CreateObject.hh>
#include <MouseMain.hh>

#include <mouse_config.h>

//
//.. constructor
//
void MObject::construct(MObject *a_parent)
{
  is_class = new TList<string> (10, 10, "");
  AddClassName();
  is_garbage = false;
  parent = a_parent;
  first_child = NULL;
  next_sibling = NULL;
  ancestor = NULL;
  num_children = 0;
  if (parent != NULL) {
    ancestor = parent->ancestor;
    mesh_id = parent->mesh_id;
    if (parent->first_child != NULL) {
      MObject *obj;
      obj = parent->first_child;
      while (obj->next_sibling != NULL) obj = obj->next_sibling;
      obj->next_sibling = this;
    } else {
      parent->first_child = this;
    };
    char new_id[200];
    strcpy(new_id, "");
    sprintf(new_id, "_%d", parent->num_children);
    id_text = parent->id_text + new_id;
    parent->num_children++;
  } else {
    id_text = "";
    mesh_id = id_text;
  };
  unique_on_mesh = false;
};  

MObject::MObject(ConstructArg arg)
{
  construct(arg.parent);
};

MObject::MObject(MObject *a_parent)
{
  construct(a_parent);
};

//
//.. DefineAsGarbage
//
void MObject::DefineAsGarbage()
{
  is_garbage = true; 
  //BroadcastEvent E("ObjectWillBeDeleted");
  //SendEvent(E);
  Ancestor()->AddToTrash(this);
};

//
//.. destructor
//
MObject::~MObject ()
{
  //cout << "deleting " << ID() << endl;
  if (is_garbage) Ancestor()->DelFromTrash(this);

  // remove object from parent's children list (if it has any)
  if (parent != NULL) {
    if (parent->first_child == this) {
      parent->first_child = next_sibling;
    } else {
      MObject *obj;
      obj = parent->first_child;
      if (obj != NULL) {
	while (obj->next_sibling != this) {
	  obj = obj->next_sibling;
	  if (obj == NULL) Bug();
	};
	obj->next_sibling = next_sibling;
      };
    };
  };

  // define kids as garbage
  MObject *obj1, *obj2;
  obj1 = first_child;
  while (obj1 != NULL) {
    obj2 = obj1->next_sibling;
    obj1->DefineAsGarbage();
    // link kids through to our dad
    if (this != ancestor) {
      obj1->parent = parent;
    };
    obj1 = obj2;
  };
  if ((first_child != NULL) && (this != ancestor)) {
    MObject *pfc = parent->first_child;
    parent->first_child = first_child;
    MObject *obj = first_child;
    while (obj->next_sibling != NULL) obj = obj->next_sibling;
    obj->next_sibling = pfc;
  };
  first_child = NULL;
};

//
//.. DefineAsUniqueOnMesh
//
void MObject::DefineAsUniqueOnMesh()
{
  string class_name = typeid(*this).name();
  if (unique_on_mesh) {
    Error(class_name + " is already unique, because of it's baseclass '" 
	  + unique_class_name + "'");
  };
  //MObject *obj = (MObject*) LookForObject(class_name, false, true);
  //MObject *obj = LookForObject<MObject>(false, true);
  unique_on_mesh = true;
  unique_class_name = class_name;
  BroadcastEvent E("Need." + class_name);
  E.only_same_mesh = true;
  SendEvent(E);
  if (E.NumOutPointers() > 0) {
    //if ((obj != NULL) && (obj != this)) {
    // start a new mesh with the parent object
    if (parent->unique_on_mesh) Error("unable to create desired mesh configuration");
    if (parent->IsA<CreateObject>()) {
      parent->parent->StartNewMesh();
    } else {
      parent->StartNewMesh();
    };
  };
};

//
//.. ReceiveEvent
//
void MObject::ReceiveEvent (Event &E)
{
  // Attention, there might be major changes in this area
  // due to a necessary calling order and future message passing options ...

  MObject *obj, *orig_carrier;
  orig_carrier = E.carrier;

  // submit event to all children
  if (!E.send_only_up && (E.current_dive_depth != E.max_dive_depth)) {
    obj = first_child;
    while ((obj != NULL) && !E.done) {
      if (obj != orig_carrier) {
	E.carrier = this;
	E.current_dive_depth++;
	obj->ReceiveEvent(E);
      };
      obj = obj->next_sibling;
    };
  };

  // handle the event yourself
  if ((orig_carrier != this) && !E.done 
      && (!E.only_same_mesh || (E.sender_mesh_id == mesh_id))) {
    if (!E.single_recipient || (E.recipient_id == id_text)) {
      HandleEvent(E);
    };
  };

  // submit event to parent
  if ((parent != NULL) && (parent != orig_carrier) && !E.done && !E.send_only_down) {
    E.carrier = this;
    if (E.current_dive_depth != 0) E.current_dive_depth--;
    parent->ReceiveEvent(E);
  };

  // mark event as done if this is the sender
  if (E.sender == this) E.MarkAsDone();
};

//
//.. SendEvent
//
int MObject::SendEvent (Event &E)
{
#ifdef MDEBUG
  {
    string text;
    text = "event '" + E.name + "' ";
    text = text + "sent by object of class '" + TopClassName() + "'";
    Mesg(text);
  }
#endif
  if (!E.invalid) {
    E.sender = this;
    E.sender_id = id_text;
    E.sender_mesh_id = mesh_id;
    E.carrier = this;
    ReceiveEvent(E);
    return true;
  };
  return false;
};

//
//.. HandleEvent
//
void MObject::HandleEvent (Event &E)
{
  if (E.name == "Ping") {
    E.React(this);
  } else {
    //long int i;
    string text;
    //_FORALL(i, is_class->) {
    for(size_t i = is_class->begin(); i < is_class->end(); i = is_class->next(i)) {
      text = "Need." + is_class->At(i);
      if (E.name == text) {
	E.React(this);
	E.AddOutPointer(this);
	break;
      };
      text = "Ping." + is_class->At(i);
      if (E.name == text) E.React(this);
      if (E.Done()) break;
    };
    if ((E.name == "LookingForGarbage") && is_garbage) {
      E.React(this);
      E.AddOutPointer(this);
    };
    if (E.name == "CommunicatorChanged") {
      E.React(this);
      string new_id_text = ancestor->ID();
      unsigned int i = 0;
      while (i < id_text.size()) {
	if (id_text[i] == '_') break;
	i++;
      };
      while (i < id_text.size()) {
	new_id_text = new_id_text + id_text[i];
	i++;
      };
      id_text = new_id_text;
    };
    if (E.name == "MeshIDChanged") {
      E.React(this);
      if (unique_on_mesh) {
	MStringList *u_classes = dynamic_cast<MStringList*>(E.InPointer(0));
	if (u_classes == NULL) {
	  string text = "problems getting event parameter u_classes\n";
	  text = text + "please send a bug report!";
	  Error(text);
	};
	for(size_t i = u_classes->begin(); i < u_classes->end(); i = u_classes->next(i)) {
	  //cout << u_classes->At(i) << endl;
	  if (u_classes->At(i) == unique_class_name) {
	    Error("unable to create desired mesh configuration");
	  };
	};
	u_classes->NewEntry() = unique_class_name;
      };
      mesh_id = E.sender->mesh_id;
    };      
  };
};

//
//.. SetID
//
int MObject::SetID (string a_id_text)
{
  MessageEvent E(a_id_text, "Ping");
  SendEvent(E);
  if (E.num_reactions == 0) {
    id_text = a_id_text;
    return true;
  };
  return false;
};

//
//.. Error
//
void MObject::Error (string text)
{
  MouseMain *mmain = FindObject<MouseMain>();
  cout << "\nERROR :-(\n"
       << "===========\n"
       << "classname : " << TopClassName() << "\n"
       << "object ID : " << ID() << "\n";
  if (mmain->IsParsingFile()) {
    cout << "file      : " << mmain->LastFile() << "\n";
    cout << "line      : " << mmain->LastLine() << "\n";
  };
  cout << "\n"
       << "the following went wrong:\n"
       << text << "\n" 
       << "sorry..." << endl;
  exit(EXIT_FAILURE);
};

//
//.. Bug
//
void MObject::Bug() const
{
  Mesg(string("sorry ... that seems to be a bug!\n") +
       string("Please submit a bug report to mouse@vug.uni-duisburg.de"));
  exit(EXIT_FAILURE);
};

//
//.. Warning
//
void MObject::Warning (string text)
{
  cout << "warning in '" << TopClassName() << "' : " << text << endl;
};

//
//.. Mesg
//
void MObject::Mesg (string text) const
{
  cout << "\nmessage:\n"
       << "=================================================================\n"
       << "classname : " << TopClassName() << "\n"
       << "object ID : " << ID() << "\n"
       << "-----------------------------------------------------------------\n"
       << text << "\n" << endl;
};

//
//.. Abstract
//
void MObject::Abstract ()
{
  Error ("Attempt to use an abstract method ==> sorry ...");
};


//
//.. GetRealValue
//
real MObject::GetRealValue(string name, string mesh)
{
  RequestEvent E("Need.ValueOnMesh");
  E.AddInString(mesh);
  E.AddInString(name);
  SendEvent(E);
  if (!E.Success()) Error("could not find '" + name + "' on the mesh'" + MeshID() + "'");
  return E.OutReal(0);
};

//
//.. GetIntValue
//
long int MObject::GetIntValue(string name, string mesh)
{
  RequestEvent E("Need.ValueOnMesh");
  E.AddInString(mesh);
  E.AddInString(name);
  SendEvent(E);
  if (!E.Success()) Error("could not find '" + name + "' on the mesh'" + MeshID() + "'");
  return E.OutInt(0);
}; 

//
//.. IsA
//
/*
bool MObject::IsA (string class_name)
{
  for(size_t i = is_class->begin(); i < is_class->end(); i = is_class->next(i)) {
    if (class_name == is_class->At(i)) return true;
  };
  return false;
};
*/
      
//
//.. StartNewMesh
//
void MObject::StartNewMesh()
{
  mesh_id = id_text;
  MStringList u_classes(ConstructArg(this, "string", ""));
  BroadcastEvent E("MeshIDChanged");
  E.send_only_down = true;
  E.AddInPointer(&u_classes);
  SendEvent(E);
};

//
//.. CreateObjectTree
//
void MObject::CreateObjectTree(ostream &s, TList<MObject*> *objects, 
			       TList<MObject*> *end_objects,
			       int max_items, int depth, int max_depth,
			       int gap_depth, string font1, string font2)
{
  long int n;
  bool bundle;
  MObject *child;
  char TAB = 9;
  if (depth == 0) {
    objects = new TList<MObject*>(100, 100, NULL);
    end_objects = new TList<MObject*>(100, 100, NULL);
    if (first_child == NULL) return;
    max_depth = 0;
    int d = 0;
    int items = 0;
    CalcNumObjectsInDepth(depth, max_depth, items);
    while ((items <= max_items) && (max_depth < 10)) {
      max_depth++;
      items = 0;
      d = 0;
      CalcNumObjectsInDepth(d, max_depth, items);
    };
    max_depth--;
    if (max_depth == 0) return;
    s << "\\setlength{\\GapDepth}{" << gap_depth << "mm}" << endl;
    s << "{\\" << font1 << endl;
  };
  if (max_depth == 0) return;
  n = objects->AddEntry();
  objects->At(n) = this;
  if ((depth != max_depth) && (first_child != NULL)) {
    s << "\\begin{bundle}{" << n+1 << "}" << endl;
    bundle = true;
  } else {
    s << n+1;
    bundle = false;
    if (first_child != NULL) end_objects->NewEntry() = this;
  };
  if (depth != max_depth) {
    child = first_child;
    while (child != NULL) {
      s << "\\chunk{";
      child->CreateObjectTree(s, objects, end_objects, max_items, depth + 1, 
			      max_depth, gap_depth, font1, font2);
      child = child->next_sibling;
      s << "}" << endl;
    };
  };
  if (bundle) {
    s << "\\end{bundle}";
  };
  if (depth == 0) {
    s << "}" << endl;
    //long int i;
    s << "\\vspace{1cm} \\\\" << endl;
    s << "{\\" << font2 << endl;
    //_FORALL(i, objects->) {
    for (size_t i=objects->begin(); i<objects->end(); i=objects->next(i)) {
      s << "\\begin{boxedminipage}[t]{\\textwidth}\n";
      s << "\\begin{verbatimtab}\n";
      s << i+1 << TAB << objects->At(i)->TopClassName() << endl;
      s << "ID:" << TAB << objects->At(i)->ID() << "\n";
      s << "mesh:" << TAB << objects->At(i)->mesh_id << "\n";
      s << "\\end{verbatimtab}\n";
      s << "\\end{boxedminipage}\n" << endl;
    };
    s << "}" << endl;
    s << "\\clearpage" << endl;
    //_FORALL(i, end_objects->) {
    for (size_t i=end_objects->begin(); i<end_objects->end(); i=end_objects->next(i)) {
      int md = 0;
      end_objects->At(i)->CreateObjectTree(s, NULL, NULL, max_items, 0, 
					   md, gap_depth, font1, font2);
    };
    delete objects;
    delete end_objects;
  };
};

//
//.. NumObjectsInDepth
//
void MObject::CalcNumObjectsInDepth(int depth, int max_depth, int &N)
{
  if ((depth == max_depth) || (first_child == NULL)) {
    N++;
  } else {
    MObject *child = first_child;
    while (child != NULL) {
      child->CalcNumObjectsInDepth(depth+1, max_depth, N);
      child = child->next_sibling;
    };
  };
};









