// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#if !defined(MOBJECT_HH)
#define MOBJECT_HH

#include <string>
#include <typeinfo>

class MObject;
class Event;
class Communicator;

#include <mouse_types.hh>
#include <TList.hh>

/**
   MObject is the basis for most of the classes within the MOUSE-project.
   It provides the functionality of sending and receiving events, as well
   as organizing instantiated objects in a tree structure.
*/
class MObject
{
  friend class ConstructArg;
  friend class Communicator;
  friend class MouseMain;

private:
  /** */ unsigned int num_children;
  /** */ MObject *first_child;
  /** */ MObject *next_sibling;
  /** */ void ReceiveEvent (Event &E);
  /** */ bool is_garbage;
  /** */ void construct(MObject *a_parent);

protected:
  /** */ bool unique_on_mesh;
  /** */ string unique_class_name;
  /** */ long int data_int;
  /** */ real data_real;
  /** */ string data_string;
  /** */ string id_text;
  /** */ string mesh_id;
  /** */ TList<string> *is_class;
  /** */ Communicator *ancestor;
  /** */ MObject *parent;
  /** */ virtual void HandleEvent (Event &E);
  /** */ int SetID (string a_id_text);
  /** */ void Abstract ();
  /** */ int SendEvent (Event &E);
  /** */ virtual void AddClassName () {
    is_class->NewEntry() = typeid(*this).name(); 
  }
  /** */ void StartNewMesh();
  /** */ virtual void DefineAsUniqueOnMesh();

public:
  /** */ MObject (ConstructArg arg);
  /** */ MObject (MObject *a_parent);
  /** */ virtual ~MObject ();

  /** */ string ID() const { return id_text; };
  /** */ string MeshID() const { return mesh_id; };
  /** */ MObject* Parent() const { return parent; };
  /** */ MObject* FirstChild() const { return first_child; };
  /** */ MObject* NextSibling() const { return next_sibling; };
  /** */ Communicator* Ancestor() const { return ancestor; };
  /** */ void Error(string text);
  /** */ void Bug() const;
  /** */ void Warning (string text);

  /** write a message */ 
  void Mesg (string text) const;

  /** 
      send a request to build an object of the type T.
      This should only be called from within a constructor.
      @param arg the ConstructArg from the constructor we are currently in
      @return a pointer to the object if it could be built
  */
  template<class T>
  T* BuildObject(ConstructArg arg);

  /** */
  template<class T>
  void BuildObject(T* &obj, ConstructArg arg) { obj = BuildObject<T>(arg); };

  /**
      look for an object of the type T
      @param only_up set this to true if you only want to
      go up in the object tree
      @param only_same_mesh set this to true if you only accept 
      objects in the same mesh
      @return a pointer to the found object, 
      or NULL if nothing was found  
  */ 
  template<class T>
  T* LookForObject (bool only_up = false,
		    bool only_same_mesh = false);

  /**
     This can be called from outside an MObject.
  */
  template<class T>
  void LookForObject(T* &obj, bool send_only_up = false, bool only_same_mesh = false) {
    obj = LookForObject<T>(send_only_up, only_same_mesh); 
  };
    

  /** */ real GetRealValue(string name, string mesh);
  /** */ long int GetIntValue(string name, string mesh);

  /**
     resolve a variable definition of the type T.
     This should only be called from within a constructor.
     @param arg the ConstructArg from the constructor we are currently in
     @return a pointer to the object belonging to the variable
  */
  template<class T>
  T* ResolveVar(ConstructArg arg);

  /**
     This can be called from outside an MObject.
  */
  template<class T>
  void ResolveVar(T* &obj, ConstructArg arg) { obj = ResolveVar<T>(arg); };

  /**
     the same as FindObject, but with an error if nothing is found
     @param only_up set this to true if you only want to
     go up in the object tree
     @param only_same_mesh set this to true if you only accept 
     objects in the same mesh
     @return a pointer to the found object
  */ 
  template<class T>
  T* FindObject (bool send_only_up = false,
		 bool only_same_mesh = false);

  /**
     This can be called from outside an MObject.
  */
  template<class T>
  void FindObject(T* &obj, bool send_only_up = false, bool only_same_mesh = false) { 
    obj = FindObject<T>(send_only_up, only_same_mesh); 
  };

  /** 
      Check if this object is an object of type T.
      @return flag if it is of type T
  */ 
  template<class T>
  bool IsA();

  /**
     Check if an object is an object of type T.
     @param obj pointer to any MObject
     @return flag if it is of type T
  */ 
  template<class T>
  bool IsA(MObject* obj);

  /** */ virtual string ClassName() const { return typeid(*this).name(); };
  /** */ string TopClassName() const { return is_class->At(is_class->NumEntries()-1); };
  
  /** mark object as garabge. */
  void DefineAsGarbage();

  /**
     check if object is garbage.
     @return true if object is marked as garbage 
  */
  bool IsGarbage() { return is_garbage; };

  /** */ void CreateObjectTree(ostream &s, TList<MObject*> *objects, 
			       TList<MObject*> *end_objects,
			       int max_items, int depth, int max_depth,
			       int gap_depth, string font1, string font2);
  /** */ void CalcNumObjectsInDepth(int depth, int max_depth, int &N);

  /** */ MObject* ThisMObject() { return this; };
  /** */ virtual void* ThisInherited() { return (void*) this; }
};

 
#include "RequestEvent.hh" 

template<class T>
T* MObject::BuildObject(ConstructArg arg)
{
  string class_name = typeid(T).name();
  //cout << "building a " << class_name << endl;
  bool var_defined = false;
  string command, var_name;

  // read in the MCL command
  command = arg.P->ReadString();
  if (command[0] == '%') {
    var_name = command;
    command = arg.P->ReadString();
    var_defined = true;
  };

  // send an Event and ask for someone building a 'T'
  RequestEvent E("Build." + command);
  E.P = arg.P->SubParser();
  SendEvent(E);
  if (!E.Success()) {
    Error("couldn't build a '" + class_name + "' from '" + command + "'");
  };

  // check if we realy got a 'T' or something else
  {
    MObject *obj = E.OutPointer(0);
    string got_class = typeid(obj).name();
    if (dynamic_cast<T*>(obj) == NULL) {
      Error("couldn't build a '" + class_name + "' from '" + command + 
	    "' got a '" + got_class + "' instead");
    };
  };
  E.P->DefineAsGarbage();

  T *obj = (T*) E.OutPointer(0);
  if (var_defined) arg.P->AddVar(var_name, obj);
  return obj;
};

template<class T>
T* MObject::LookForObject(bool send_only_up,
			  bool only_same_mesh)
{
  string class_name = typeid(T).name();
  RequestEvent E("Need." + class_name);
  E.send_only_up = send_only_up;
  E.only_same_mesh = only_same_mesh;
  SendEvent(E);
  if (!E.Success()) return NULL;
  T* obj = dynamic_cast<T*>(E.OutPointer(0));
  if (obj == NULL) {
    Error("got wrong type from 'Need...' event for class '" + class_name + "'");
  };
  return obj;
};

template<class T>
T* MObject::ResolveVar(ConstructArg arg)
{
  string class_name = typeid(T).name();
  MObject *obj_M = arg.P->ResolveVar();
  T *obj_T = dynamic_cast<T*>(obj_M);
  if (obj_T == NULL) {
    Error("could not resolve variable of type " + class_name);
  };
  return (T*) obj_T;
};

template<class T>
T* MObject::FindObject(bool send_only_up,
		       bool only_same_mesh)
{
  string class_name = typeid(T).name();
  T *obj;
  obj = LookForObject<T>(send_only_up, only_same_mesh);
  if (obj == NULL) {
    Error("couldn't find object of class '" + class_name + "'");
  };
  return obj;
};

template<class T>
bool MObject::IsA()
{
  if (dynamic_cast<T*>(this) != NULL) return true;
  return false;
};

template<class T>
bool MObject::IsA(MObject *obj)
{
  if (dynamic_cast<T*>(obj) != NULL) return true;
  return false;
};
  


#endif















