// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include <mouse_config.h>
#ifdef WITH_QT

#include "MQDrawWidget.hh"
#include <ConstructArg.hh>
#include <geometry2D.hh>

//
//.. constructor
//
MQDrawWidget::MQDrawWidget(ConstructArg arg, QWidget *parent, 
			   const char *name, WFlags f)
  : MQWidget(arg, parent, name, f)
{
  AddClassName();
  PL = FindObject<PointList>(false, true);
  EL = FindObject<ElementList>(false, true);
  setBackgroundColor(white);
  first_paint = true;
  is_painting = false;
  other_scale =false;
};

//
//.. WorldToScreen
//
void MQDrawWidget::WorldToScreen(real xw, real yw, int &xs, int &ys)
{
  real dxw = xw - x_origin;
  real dyw = yw - y_origin;
  if (!other_scale) {
    xs = (int) ((dxw / x_screen) * width());
    ys = height() - (int) ((dyw / y_screen) * height());
  } else {
    xs = (int) ((dxw / x_screen) * scale_rect.width());
    ys = scale_rect.height() - (int) ((dyw / y_screen) * scale_rect.height());
  };
};

//
//.. ScreenToWorld
//
void MQDrawWidget::ScreenToWorld(int xs, int ys, real &xw, real &yw)
{
  if (!other_scale) {
    xw = x_origin + ((real) xs) / width() * x_screen;
    yw = y_origin + ((real) (height() - ys)) / height() * y_screen;
  } else {
    xw = x_origin + ((real) xs) / scale_rect.width() * x_screen;
    yw = y_origin + ((real) (scale_rect.height() - ys)) / scale_rect.height() * y_screen;
  };
};

//
//.. Zoom
//
void MQDrawWidget::Zoom(real xz, real yz)
{
  real xm = x_origin + .5*x_screen;
  real ym = y_origin + .5*y_screen;
  x_screen -= .01 * xz * x_screen;
  y_screen -= .01 * yz * y_screen;
  x_origin = xm - .5 * x_screen;
  y_origin = ym - .5 * y_screen;
  repaint();
};

//
//.. resizeEvent()
//
void MQDrawWidget::resizeEvent(QResizeEvent *E)
{
  OptimizeAspectRatio();
};

void MQDrawWidget::OptimizeAspectRatio()
{
  if ((width() > 0) && (height() > 0)) {
    real scale = ((real)height())/((real)width());
    if (scale > 1) {
      real ym = y_origin + .5 * y_screen;;
      y_screen = x_screen * scale;
      y_origin = ym - .5*y_screen;
    } else {
      scale = 1.0/scale;
      real xm = x_origin + .5 * x_screen;;
      x_screen = y_screen * scale;
      x_origin = xm - .5*x_screen;
    };
  };
};

//
//.. Autoscale
//
void MQDrawWidget::Autoscale()
{
  bool first = true;
  real x_min=0, x_max=0, y_min=0, y_max=0;
  FORALL(i, PL->) {
    if (first) {
      x_max = x_min = PL->x(i);
      y_max = y_min = PL->y(i);
      first = false;
    } else {
      if (PL->x(i) < x_min) x_min = PL->x(i);
      if (PL->y(i) < y_min) y_min = PL->y(i);
      if (PL->x(i) > x_max) x_max = PL->x(i);
      if (PL->y(i) > y_max) y_max = PL->y(i);
    };
  };
  x_min = x_min - .1*(x_max-x_min);
  x_max = x_max + .1*(x_max-x_min);
  y_min = y_min - .1*(y_max-y_min);
  y_max = y_max + .1*(y_max-y_min);
  x_screen = x_max - x_min;
  y_screen = y_max - y_min;
  x_origin = x_min;
  y_origin = y_min;
  OptimizeAspectRatio();
  repaint();
};

//
//.. Move
//
void MQDrawWidget::Move(real dx, real dy)
{
  x_origin += dx * x_screen;
  y_origin += dy * y_screen;
  repaint();
};  

//
//.. DrawLine
//
void MQDrawWidget::DrawLine(QPainter *P, real x1, real y1, real x2, real y2)
{
  int xi1, yi1, xi2, yi2;
  WorldToScreen(x1, y1, xi1, yi1);
  WorldToScreen(x2, y2, xi2, yi2);
  P->drawLine(xi1, yi1, xi2, yi2);
};

void MQDrawWidget::DrawLine(QPainter *P,size_t p1, size_t p2)
{
  DrawLine(P, PL->x(p1), PL->y(p1), PL->x(p2), PL->y(p2));
};

//
//.. DrawVector
//
void MQDrawWidget::DrawVector(QPainter *P, real x1, real y1, real vx, real vy, 
			      real size, real head_size)
{
  real x2 = x1 + size*vx;
  real y2 = y1 + size*vy;
  real nx = vx*size*head_size;
  real ny = vy*size*head_size;
  TurnLeft(nx, ny);
  real x3 = x1 + size*vx - vx*size*head_size + nx;
  real y3 = y1 + size*vy - vy*size*head_size + ny;
  real x4 = x1 + size*vx - vx*size*head_size - nx;
  real y4 = y1 + size*vy - vy*size*head_size - ny;
  int xi1, xi2, xi3, xi4, yi1, yi2, yi3, yi4;
  WorldToScreen(x1, y1, xi1, yi1);
  WorldToScreen(x2, y2, xi2, yi2);
  WorldToScreen(x3, y3, xi3, yi3);
  WorldToScreen(x4, y4, xi4, yi4);
  P->drawLine(xi1, yi1, xi2, yi2);
  P->drawLine(xi2, yi2, xi3, yi3);
  P->drawLine(xi2, yi2, xi4, yi4);
};

void MQDrawWidget::DrawCircle(QPainter *P, real x, real y, real r)
{
  int x_min, y_min, x_max, y_max;
  WorldToScreen(x-r, y-r, x_min, y_min);
  WorldToScreen(x+r, y+r, x_max, y_max);
  int w = x_max - x_min;
  P->drawEllipse(x_min, y_max, w, w);
};


#endif














