// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#if !defined(MESH2D_HH)
#define MESH2D_HH

class Mesh2D;

#include <MObject.hh>
#include <PointList.hh>
#include <ElementList.hh>
#include <PelList2D.hh>
#include <Triangulator2D.hh>

/**
   Describes a twodimensional mesh. This is used for mesh generation.
   A general base class for meshes might be written in the future. 
*/
class Mesh2D : public MObject
{
protected:
  /** 
      1,1 component of stretching matrix
  */ TList<real> *a11;

  /** 
      1,2 component of stretching matrix
  */ TList<real> *a12;

  /** 
      2,2 component of stretching matrix
  */ TList<real> *a22;

  /** 
      net density value
  */ TList<real> *nd;

  /** 
      flag to determin if a point may be altered or not
  */ TList<bool> *fixed;

  /** 
      flag to mark points for modifications
  */ TList<bool> *marked;

  /** 
      point 1 of an initial boundary edge
  */ //TList<size_t> *front_p1;

  /** 
      point 1 of an initial boundary edge
  */ //TList<size_t> *front_p2;

  /** 
      the point information 
  */ PointList *PL;

  /** 
      the element information
  */ ElementList *EL;

  /** 
      additional information which elements are connected 
      to which points
  */ PelList2D *PEL;

  /** 
      an initial raising bubble triangulator
  */ 
  Triangulator2D *TRI;

public:
  /** */ Mesh2D (ConstructArg arg);
  /** */ ~Mesh2D();

  /** 
      upadate point to element information
  */ 
  void UpdatePEL();

  /** 
      Check if an element is orientated counter clockwise.
      CheckOrientation will also change the element's orientation
      if it id clockwise.
      @param i_element the index of the element to check
      @return a flag if the orientation was counter clockwise
  */ 
  bool CheckOrientation(size_t i_element);

  /** 
      write point and element data to their specified files
  */ void WriteMeshData();

  /** 
      Find the node number of a point in an element.
      This method will throw a NotFound_error if it is not able to
      find the node.
      @param element the index of the element
      @param point the index of the point
      @return the node number of that point in this element
  */
  size_t FindNode(size_t element, size_t point);

  /** 
      Transform coordinates into the stretched space.
      @param p0 the index of the transformation origin
      @param x the x coordinate to transform
      @param y the y coordinate to transform
  */ 
  void Transform(size_t p0, real &x, real &y);

  /** 
      Transform coordinates into the stretched space.
      @param x0 the x coordinate of the transformation origin
      @param y0 the y coordinate of the transformation origin
      @param a11 the 1,1 component of the local stretching matrix
      @param a12 the 1,2 component of the local stretching matrix
      @param a22 the 2,2 component of the local stretching matrix
      @param x the x coordinate to transform
      @param y the y coordinate to transform
  */ 
  void Transform(real x0, real y0, 
		 real a11, real a12, real a22,
		 real &x, real &y);

  /** 
      Retransform coordinates to the original space.
      @param p0 the index of the transformation origin
      @param x the x coordinate to retransform
      @param y the y coordinate to retransform
  */ 
  void Retransform(size_t p0, real &x, real &y);

  /** 
      Retransform coordinates to the original space.
      @param x0 the x coordinate of the transformation origin
      @param y0 the y coordinate of the transformation origin
      @param a11 the 1,1 component of the local stretching matrix
      @param a12 the 1,2 component of the local stretching matrix
      @param a22 the 2,2 component of the local stretching matrix
      @param x the x coordinate to retransform
      @param y the y coordinate to retransform
  */ 
  void Retransform(real x0, real y0, 
		   real a11, real a12, real a22,
		   real &x, real &y);

  /** 
      Compute the eigenvalues of the stretching matrix
      @param a11 the 1,1 component of the local stretching matrix
      @param a12 the 1,2 component of the local stretching matrix
      @param a22 the 2,2 component of the local stretching matrix
      @param l1 the first eigenvalue
      @param l2 the second eigenvalue
  */ 
  void CalcEigenvalues(real a11, real a12, real a22, real &l1, real &l2);

  /** ??? */ void CalcDirVector(size_t p, real &vx, real &vy);

  /** 
      Add a new boundary segment.
      @param p1 first point of the new segment
      @param p2 second point of the new segment
  */ 
  //void AddFront(size_t p1, size_t p2);

  /** 
      Set the stretching matrix for a point.
      @param point the index of the point
      @param R1 first stretching factor
      @param R2 second stretching factor
      @param gamma angle between x-axis and the first normal direction
      of the matrix
  */ 
  void SetMatrix(size_t point, real R1, real R2, real gamma);

  /** 
      find elements connected to an edge.
      @param p1 starting point of the edge
      @param p2 ending point of the edge
      @param e1 element left of the edge
      @param e2 element right of the edge
      @param f1 flag if left element was found
      @param f2 flag if right element was found
  */
  void FindElements(size_t p1, size_t p2, size_t &e1, size_t &e2, bool &f1, bool &f2);

  /** 
      Check if two elements are neighbours.
      @return flag if they are
  */ 
  bool AreNeighbours(size_t e1, size_t e2);

  /** 
      Clean up all lists belonging to this mesh.
  */ 
  void CleanUp();

  /** 
      Check if a point can be deleted by pulling it to another one.
      @param p0 the point to delete
      @param p_pullto drag it to this one
      @return flag if possible
  */ 
  bool CheckPoint(size_t p0, size_t p_pullto);

  /** 
      Check if a point can be moved.
      @param point the point to move
      @param x_new the new x coordinate of the point
      @param y_new the new y coordinate of the point
      @param ne number of elements to exclude from the consideration (0,1,2)
      @param excl_elem1 first element to exclude
      @param excl_elem2 second element to exclude
      @return flag if possible
  */
  bool CheckPoint(size_t point, real x_new, real y_new, int ne = 0,
		  size_t excl_elem1 = 0, size_t excl_elem2 = 0);

  /**
     Get the net density for a point.
     @param i the point index
     @return the net density for point i
  */
  real ND(size_t i) { return nd->At(i); };

  /**
     Get the net 1,1 component of the stretching matrix for a point.
     @param i the point index
     @return the 1,1 component for point i
  */
  real A11(size_t i) { return a11->At(i); };

  /**
     Get the net 1,2 component of the stretching matrix for a point.
     @param i the point index
     @return the 1,2 component for point i
  */
  real A12(size_t i) { return a12->At(i); };

  /**
     Get the net 2,2 component of the stretching matrix for a point.
     @param i the point index
     @return the 2,2 component for point i
  */
  real A22(size_t i) { return a22->At(i); };

  /**
     Set the stretching matrix for a given point.
     @param i the point index
     @param a_11 the 1,1 component of the stretching matrix
     @param a_12 the 1,2 component of the stretching matrix
     @param a_22 the 2,2 component of the stretching matrix
  */
  void SetMatrixComp(size_t i, real a_11, real a_12, real a_22);

  /**
     Set the net density for a given point.
     @param i the point index
     @param net_dens the net density value for point i
  */
  void SetNetDensity(size_t i, real net_dens) { nd->At(i) = net_dens; };

  /**
     Mark a point as unmovable.
     @param i the index of the point
  */
  void FixPoint(size_t i) { fixed->At(i) = true; };

  /**
     Mark a point as movable.
     @param i the index of the point
  */
  void ReleasePoint(size_t i) { fixed->At(i) = false; };

  /**
     Check if a point is fixed.
     @param i the index of the point
     @return flag if it is fixed
  */
  bool IsFixed(size_t i) { return fixed->At(i); };

  /**
     Check if a point is marked for the next mesh operation.
     @param i the index of the point
     @return flag if it is marked
  */
  bool IsMarked(size_t i) { return marked->At(i); };

  /**
     Mark a point for the next mesh operation.
     @param i the index of the point
  */
  void MarkPoint(size_t i) { marked->At(i) = true; };

  /**
     Unmark a point for the next mesh operation.
     @param i the index of the point
  */
  void UnmarkPoint(size_t i) { marked->At(i) = false; };





  // ... more comments

  /** */ void UnionPoints(size_t p, size_t p_pullto);
  /** */ size_t FindOutsideElement(size_t p);
  /** */ int NumTriangles(size_t p);
  /** */ int NumQuads(size_t p);
  /** */ void CalcBaseMatrix (real &g11, real &g12, real &g21, real &g22,
			      real a11, real a12, real a22);
  /** */ void EigenProperties (real a11, real a12, real a22,
			       real &lambda_big, real &lambda_small,
			       real &eigen_big_vx, real &eigen_big_vy,
			       real &eigen_small_vx, real &eigen_small_vy);
  /** */ void MarkAll();
  /** */ void UnmarkAll();
  /** */ bool PointMarked(size_t p) { return marked->At(p); };
  /** */ bool ElementMarked(size_t e);
  /** */ void MarkPoints(TList<size_t> *point, int depth);
  /** */ void DeleteElement(size_t e);
  /** */ size_t AddTri(size_t p0, size_t p1, size_t p2);
  /** */ size_t AddQuad(size_t p0, size_t p1, size_t p2, size_t p3);
};

#endif
















