// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#if !defined(POINTLIST_HH)
#define POINTLIST_HH

class PointList;

#include <TList.hh>
#include <MeshDataList.hh>
#include <MIntList.hh>

/**
   PointList is used to store the metric of a mesh. This could be a structured
   curvilinear or an unstructured mesh. However PointList only stores the 
   coordinates of the points, but no structure or neighbourship information.
   \\ \\
   {\bf MCL syntax:}
   \begin{verbatim}
   PointList {
     <--
     bool read_points
   }
   \end{verbatim}
*/
class PointList : public MeshDataList 
{
  friend class ElementList;

protected:
  /** the x-coordinates of the points. */
  TList<real> *x_point;

  /** the y-coordinates of the points. */
  TList<real> *y_point;

  /** the z-coordinates of the points. */
  TList<real> *z_point;

  /** the encoding of the points */
  TList<int> *code_point;

  /** list of all boundary points */
  TList<size_t> *b_point;

  /** the x component of the boundary normal vector */
  TList<real> *b_nx;

  /** the y component of the boundary normal vector */
  TList<real> *b_ny;

  /** the z component of the boundary normal vector */
  TList<real> *b_nz;

  /** characteristic length for time step restrictions */
  TList<real> *ch_dist;

  /** inverse control volumes */
  TList<real> *ivol;

  /** field of flags to determine id the volume has been set already */
  TList<bool> *volset;

  /** read a point from a stream */
  void ReadPoint (size_t i, istream &s);

  /** write a point to a stream */
  virtual void WritePoint (size_t i, ostream &s);

public:
  /** filename for the metric. */
  string points_filename;

  /** constructor. */
  PointList (ConstructArg arg);

  /** destructor. */
  ~PointList ();

  /** Create the list of boundary points. */
  //void CreateBoundaryList();

  /** Write points to a file.
      Write all points to the file specified in the control-file. */
  void WriteAllPoints ();

  /** Read points from a file.
      Read all points from the file specified in the control-file. */
  void ReadAllPoints ();

  /** Access-method to the x-coordinate.
      @param i the index of the point
      @return a reference to the point's x-coordinate */
  real& x(size_t i) { return x_point->At(i); };

  /** Access-method to the y-coordinate.
      @param i the index of the point
      @return a reference to the point's y-coordinate */
  real& y(size_t i) { return y_point->At(i); };

  /** Access-method to the z-coordinate.
      @param i the index of the point
      @return a reference to the point's z-coordinate */
  real& z(size_t i) { return z_point->At(i); };

  /** Access-method to the boundary-code.
      @param i the index of the point
      @return a reference to the point's boundary-code */
  int& code(size_t i) { return code_point->At(i); };

  /** Find out if point belongs to a domain border.
      @param i the index of the point
      @return flag if point i is a boundary point */
  bool IsBoundary(size_t i) { return code(i) != 0; };

  /** output operator for this class ??? */
  friend ostream& operator<< (ostream&,PointList&);

  /** Check for undefined point codes
      @param sw specifies if the program should be stopped if undefined codes are found
      @return a flag indicating if undefined codes were found */
  bool FindUndefBC(bool sw);

  /** Get the corresponding field index of a boundary index.
      @param i the index of the boundary point
      @return the index of the field point */
  size_t BPoint(size_t i) { return b_point->At(i); };

  /** Get the inverse control volume of a point.
      @param i the index of the point
      @return the inverse control volume */
  real IVol(size_t i) { return ivol->At(i); };

  /** Get the control volume of a point.
      @param i the index of the point
      @return the control volume */
  real Vol(size_t i) { return 1/ivol->At(i); };

  /** Get the characteristic length of a point.
      @param i the index of the point
      @return the characteristic length */
  real ChDist(size_t i) { return ch_dist->At(i); };

  /** */ virtual void CopyEntry (size_t src, size_t dst);
  /** */ virtual void Extend (size_t delta);
  /** */ void SetIVol(size_t i, real an_ivol) { ivol->At(i) = an_ivol; };
  /** */ void SetChDist(size_t i, real a_ch_dist) { ch_dist->At(i) = a_ch_dist; };
  /** */ void SetVolSet(size_t i, bool a_volset) { volset->At(i) = a_volset; };
  /** */ bool VolSet(size_t i) { return volset->At(i); };
};

#endif



















