// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#if !defined(PopInterface2D_HH)
#define PopInterface2D_HH

class PopInterface2D;

/** This is just the definition of an interface and thus PopInterface2D will never be
    inherited. It defines which methods have to be provided for
    a class implementing a POP interface in two dimensions. A class
    implementing this interface will then be passed as template argument
    to other classes needing access to the point to other point data. */
class PopInterface2D
{
public:
  /** Get the number of field edges connected to a point. Please note that every edge
      is a field edge, but some edges might be boundary edges as well.
      @param i the index of the point
      @return the number of field edges connected to point i */ 
  size_t NumFLinks(size_t i);

  /** Get the number of boundary edges connected to a point. 
      @param i the index of the point
      @return the number of boundary edges connected to point i */ 
  size_t NumBLinks(size_t i);

  /** Get the point address of a point on the other end of a linked field edge.
      @param i the index of the point
      @param j the number of the link
      @return the address of the other point. */
  size_t FLinkPoint(size_t i, size_t j);
  
  /** Get the point address of a point on the other end of a linked boundary edge.
      @param i the index of the point
      @param j the number of the link
      @return the address of the other point. */
  size_t BLinkPoint(size_t i, size_t j);

  /** Get the x component of the interface vector of a linked field edge.
      @param i the index of the point
      @param j the number of the link
      @return the x component of the interface vector. */
  real Fnx(size_t i, size_t j);

  /** Get the y component of the interface vector of a linked field edge.
      @param i the index of the point
      @param j the number of the link
      @return the y component of the interface vector. */
  real Fny(size_t i, size_t j);

  /** Get the x component of the interface vector of a linked boundary edge.
      Please note that for 2D boundary edges the cell interface is lying on the edge
      itself.
      @param i the index of the point
      @param j the number of the link
      @return the x component of the interface vector. */
  real Bnx(size_t i, size_t j);

  /** Get the y component of the interface vector of a linked boundary edge.
      Please note that for 2D boundary edges the cell interface is lying on the edge
      itself.
      @param i the index of the point
      @param j the number of the link
      @return the y component of the interface vector. */
  real Bny(size_t i, size_t j);

  /** Get the group ID of a point. For parallel computations the computational domain
      has to be subdivided into several groups according to the number of CPUs used.
      @param i the index of the point
      @return the group ID of the point */ 
  size_t GroupID(size_t i);

  /**
     Add a field link. 
     Please be very careful with this method, unless you really
     know what you are doing. This will alter the mesh connectivity and is
     not necessary in normal operation.
     @param p1 the starting point of the edge
     @param p2 the ending point of the edge
     @param nx x component of the interface vector
     @param ny y component of the interface vector
     (method added in version 1.0.1) 
     (method might still change)
  */
  void AddFLink(size_t p1, size_t p2, real nx, real ny);

  /**
     Add a boundary link. 
     Please be very careful with this method, unless you really
     know what you are doing. This will alter the mesh connectivity and is
     not necessary in normal operation.
     The boundary vector will point left of the edge, meaning it will be
     the vector .5*[x2-x1,y2-y1] turned left by 90 degress. So it will be
     .5*[y1-y2,x2-x1].
     @param p1 the starting point of the edge
     @param p2 the ending point of the edge 
     (method added in version 1.0.1) 
     (method might still change)
  */
  void AddBLink(size_t p1, size_t p2);

  /**
     Find the link number of a field edge.
     FLinkPoint(p1, FindFLink(p1,p2)) == p2 !!
     @param p1 the starting point of the edge
     @param p2 the ending point of the edge
     @return the link number for the point p1
     This method will throw a NotFound_error if the edge p1->p2 
     does not exists. 
     (method added in version 1.0.1) 
     (method might still change)
  */
  size_t FindFLink(size_t p1, size_t p2);

  /**
     Find the link number of a boundary edge.
     BLinkPoint(p1, FindBLink(p1,p2)) == p2 !!
     @param p1 the starting point of the edge
     @param p2 the ending point of the edge
     @return the link number for the point p1
     This method will throw a NotFound_error if the edge p1->p2 
     does not exists. 
     (method added in version 1.0.1) 
     (method might still change)
  */
  size_t FindBLink(size_t p1, size_t p2);

  /**
     Set the x component normal vector of a field link to a given value.
     Please be very careful with this method, unless you really
     know what you are doing. This will alter the mesh connectivity and is
     not necessary in normal operation.
     @param i the index of the starting point
     @param j the number of the link
     @param the x component of the normal vector
     (method added in version 1.0.1) 
     (method might still change)
  */
  void SetFnx(size_t i, size_t j, real nx);

  /**
     Set the y component normal vector of a field link to a given value.
     Please be very careful with this method, unless you really
     know what you are doing. This will alter the mesh connectivity and is
     not necessary in normal operation.
     @param i the index of the starting point
     @param j the number of the link
     @param the y component of the normal vector
     (method added in version 1.0.1) 
     (method might still change)
  */
  void SetFny(size_t i, size_t j, real ny);

  /**
     Set the x component normal vector of a boundary link to a given value.
     Please be very careful with this method, unless you really
     know what you are doing. This will alter the mesh connectivity and is
     not necessary in normal operation.
     @param i the index of the starting point
     @param j the number of the link
     @param the x component of the normal vector
     (method added in version 1.0.1) 
     (method might still change)
  */
  void SetBnx(size_t i, size_t j, real nx);

  /**
     Set the y component normal vector of a boundary link to a given value.
     Please be very careful with this method, unless you really
     know what you are doing. This will alter the mesh connectivity and is
     not necessary in normal operation.
     @param i the index of the starting point
     @param j the number of the link
     @param the y component of the normal vector
     (method added in version 1.0.1) 
     (method might still change)
  */
  void SetBny(size_t i, size_t j, real ny);
};


#endif







