// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include "PopList2D.hh"
#include <ConstructArg.hh>
#include <geometry2D.hh>
#include <POperation.hh>
#include <TPOperation.hh>
#include <TBPOperation.hh>
#include <TFEOperation2D.hh>

PopList2D::PopList2D(ConstructArg arg) : MObject(arg)
{
  AddClassName();
  PL = FindObject<PointList>(false, true);
  BPL = FindObject<BoundaryPointList>(false, true);
  EL = FindObject<ElementList>(false, true);
  flink = new TList<TList<link_t>*>(PL, NULL);
  FORALL(i, PL->) {
    flink->At(i) = new TList<link_t>(1,1);
  };
  blink = new TList<TList<link_t>*>(BPL, NULL);
  FORALL(i, BPL->) {
    blink->At(i) = new TList<link_t>(1,1);
  };
  group_id = new TList<size_t>(PL, 0);
  Update();
};

void PopList2D::DelAllSublists()
{
  FORALL(i, flink->) {
    flink->At(i)->DelAll();
    flink->At(i)->CleanUp();
  };
  FORALL(i, blink->) {
    blink->At(i)->DelAll();
    blink->At(i)->CleanUp();
  };
};

void PopList2D::reset_op::operate(size_t i, size_t p)
{
  PL->SetIVol(p, 0);
  PL->SetChDist(p, 0);
  PL->SetVolSet(p, false);
};

void PopList2D::calcvol_op::operate(size_t ie, size_t ip, size_t from, size_t to, 
				    real nx, real ny, 
				    bool mod_from, bool mod_to)
{
  // vector leading along the edge
  ax = PL->x(to) - PL->x(from);
  ay = PL->y(to) - PL->y(from);
  // sqare of "area" of control interface (absolute value of ndA)
  qarea = (nx * nx + ny * ny);
  // normal distance to cell interface
  zylvol = (ax * nx + ay * ny);
  i_pro_dist = sqrt(qarea)/zylvol*2.;
  // minimize the distance for points
  if (mod_to) {
    if(PL->ChDist(to) < i_pro_dist) PL->SetChDist(to, i_pro_dist);
  };
  if (mod_from) {
    if(PL->ChDist(from) < i_pro_dist) PL->SetChDist(from, i_pro_dist);
  };
  // add partial volumes to control-volumes
  if (mod_to) PL->SetIVol(to, PL->IVol(to) + .25*zylvol);
  if (mod_from) PL->SetIVol(from, PL->IVol(from) + .25*zylvol);
  // flag to protect dummy points
  if (mod_to) PL->SetVolSet(to, true);
  if (mod_from) PL->SetVolSet(from, true);
};

void PopList2D::postop_op::operate(size_t i, size_t p)
{
  // Ensure diviseability (dirty programming)
  if (!PL->VolSet(p)) {
    PL->SetIVol(p, 1);
    PL->SetChDist(p, 1);
  };
  // Compute inverse volume
  PL->SetIVol(p, 1/PL->IVol(p));
  PL->SetChDist(p, 1/PL->ChDist(p));
};

PopList2D::normvec_op::normvec_op(carg_t arg) : Operator(arg)
{
  AddClassName();
  BPL = FindObject<BoundaryPointList>(false, true);
  POP = FindObject<PopList2D>(false, true);
};

void PopList2D::normvec_op::operate(size_t gi, size_t i)
{
  POP->CalcNVec(i);
};

void PopList2D::CalcNVec(size_t i)
{
  real nx = 0;
  real ny = 0;
  for (size_t j = 0; j < NumBLinks(i); j++) {
    if (BPL->Prio(i) <= BPL->Prio(BLinkPoint(i,j))) {
      nx += Bnx(i,j);
      ny += Bny(i,j);
    };
  };
  VNORM2(nx,ny);
  BPL->SetNx(i,nx);
  BPL->SetNy(i,ny);
};

void PopList2D::Update()
{
  FORALL(i_element, EL->) {
    for (size_t i_node = 0; i_node < EL->NumNodes(i_element); i_node++) {
      size_t p1 = EL->PermNode(i_element, i_node);
      size_t p2 = EL->PermNode(i_element, i_node + 1);
      if (p1 > p2) {
	size_t h = p1;
	p1 = p2;
	p2 = h;
      };
      try {
	flink->At(p1)->FindItem(p2);
      }
      catch (NotFound_error) {
	size_t i_link = flink->At(p1)->AddEntry();
	flink->At(p1)->At(i_link).count = 0;
	flink->At(p1)->At(i_link).point = p2;
	flink->At(p1)->At(i_link).nx = 0;
	flink->At(p1)->At(i_link).ny = 0;
      };
    };
  };
  FORALL(i_element, EL->) {
    for (size_t i_node = 0; i_node < EL->NumNodes(i_element); i_node++) {
      size_t p1 = EL->PermNode(i_element, i_node);
      size_t p2 = EL->PermNode(i_element, i_node + 1);
      bool left = true;
      if (p1 > p2) {
	size_t h = p1;
	p1 = p2;
	p2 = h;
	left = false;
      };
      size_t i_link = flink->At(p1)->FindItem(p2);
      flink->At(p1)->At(i_link).count++;
      real nx = .5*(PL->x(p1) + PL->x(p2)) - EL->x_middle(i_element);
      real ny = .5*(PL->y(p1) + PL->y(p2)) - EL->y_middle(i_element);
      if (left) TurnLeft(nx, ny);
      else TurnRight(nx, ny);
      flink->At(p1)->At(i_link).nx += nx;
      flink->At(p1)->At(i_link).ny += ny;
    };
  };
  FORALL(i_point, flink->) {
    FORALL(i_link, flink->At(i_point)->) {
      link_t &link = flink->At(i_point)->At(i_link);
      try {
	flink->At(link.point)->FindItem(i_point);
      }
      catch (NotFound_error) {
	link_t new_link;
	new_link.nx = link.nx;
	new_link.ny = link.ny;
	new_link.point = i_point;
	new_link.count = link.count;
	flink->At(link.point)->NewEntry() = new_link;
      };
    };
  };
  FORALL(i_element, EL->) {
    for (size_t i_node = 0; i_node < EL->NumNodes(i_element); i_node++) {
      size_t p1 = EL->PermNode(i_element, i_node);
      size_t p2 = EL->PermNode(i_element, i_node + 1);
      if ((PL->code(p1) != 0) && (PL->code(p2) != 0)) {
	size_t i_link = flink->At(p1)->FindItem(p2);
	link_t f_link = flink->At(p1)->At(i_link);
	if (f_link.count == 1) {
	  {
	    link_t &b_link = blink->At(BPL->BIndex(p1))->NewEntry();
	    b_link.nx = .5*(PL->x(p2) - PL->x(p1));
	    b_link.ny = .5*(PL->y(p2) - PL->y(p1));
	    TurnLeft(b_link.nx, b_link.ny);
	    b_link.point = p2;
	  };
	  {
	    link_t &b_link = blink->At(BPL->BIndex(p2))->NewEntry();
	    b_link.nx = .5*(PL->x(p2) - PL->x(p1));
	    b_link.ny = .5*(PL->y(p2) - PL->y(p1));
	    TurnLeft(b_link.nx, b_link.ny);
	    b_link.point = p1;
	  };
	};
      };
    };    
  };

  UpdateMetric();

  {
    TBPOperation<normvec_op, PopList2D> nvec(this);
    for(size_t i = 0; i < BPL->NumCodes(); i++) nvec.AddCode(BPL->Code(i));
    nvec.Act();
  };

};

void PopList2D::UpdateMetric()
{
  TPOperation<reset_op, PopList2D> reset(this);
  TFEOperation2D<calcvol_op, PopList2D> calc(this);
  TPOperation<postop_op, PopList2D> post(this);
  reset.Act();
  calc.Act();
  post.Act();
};

void PopList2D::AddFLink(size_t p1, size_t p2, real nx, real ny)
{
  {
    if (flink->At(p1) == NULL) {
      flink->At(p1) = new TList<link_t>(1,1);
    };
    for (size_t i = 0; i < NumFLinks(p1); i++) {
      if (FLinkPoint(p1, i) == p2) {
	Error("the edge you want to add exists already");
      };
    };
    link_t &link = flink->At(p1)->NewEntry();
    link.point = p2;
    link.nx = nx;
    link.ny = ny;
  };
  {
    if (flink->At(p2) == NULL) {
      flink->At(p2) = new TList<link_t>(1,1);
    };
    for (size_t i = 0; i < NumFLinks(p2); i++) {
      if (FLinkPoint(p2, i) == p1) {
	Error("the edge you want to add exists already");
      };
    };
    link_t &link = flink->At(p2)->NewEntry();
    link.point = p1;
    link.nx = nx;
    link.ny = ny;
  };
};

void PopList2D::AddBLink(size_t p1, size_t p2, real nx, real ny)
{
  size_t bp1 = BPL->BIndex(p1);
  size_t bp2 = BPL->BIndex(p2);
  {
    if (blink->At(bp1) == NULL) {
      blink->At(bp1) = new TList<link_t>(1,1);
    };
    link_t &link = blink->At(bp1)->NewEntry();
    link.point = p2;
    link.nx = nx;
    link.ny = ny;
  };
  {
    if (blink->At(bp2) == NULL) {
      blink->At(bp2) = new TList<link_t>(1,1);
    };
    link_t &link = blink->At(bp2)->NewEntry();
    link.point = p1;
    link.nx = nx;
    link.ny = ny;
  };
};

void PopList2D::AddBLink(size_t p1, size_t p2)
{
  size_t bp1 = BPL->BIndex(p1);
  size_t bp2 = BPL->BIndex(p2);
  real nx = .5*(PL->x(p2) - PL->x(p1));
  real ny = .5*(PL->y(p2) - PL->y(p1));
  TurnLeft(nx, ny);
  {
    if (blink->At(bp1) == NULL) {
      blink->At(bp1) = new TList<link_t>(1,1);
    };
    for (size_t i = 0; i < NumBLinks(bp1); i++) {
      if (BLinkPoint(bp1, i) == p2) {
	Error("the edge you want to add exists already");
      };
    };
    link_t &link = blink->At(bp1)->NewEntry();
    link.point = p2;
    link.nx = nx;
    link.ny = ny;
  };
  {
    if (blink->At(bp2) == NULL) {
      blink->At(bp2) = new TList<link_t>(1,1);
    };
    for (size_t i = 0; i < NumBLinks(bp2); i++) {
      if (BLinkPoint(bp2, i) == p1) {
	Error("the edge you want to add exists already");
      };
    };
    link_t &link = blink->At(bp2)->NewEntry();
    link.point = p1;
    link.nx = nx;
    link.ny = ny;
  };
  CalcNVec(p1);
  CalcNVec(p2);
};

size_t PopList2D::FindFLink(size_t p1, size_t p2)
{
  if (flink->At(p1) == NULL) throw NotFound_error();
  for (size_t i = 0; i < NumFLinks(p1); i++) {
    if (FLinkPoint(p1, i) == p2) return i;
  };
  throw NotFound_error();
  return 0;
};

size_t PopList2D::FindBLink(size_t p1, size_t p2)
{
  if (blink->At(BPL->BIndex(p1)) == NULL) throw NotFound_error();
  for (size_t i = 0; i < NumBLinks(p1); i++) {
    if (BLinkPoint(p1, i) == p2) return i;
  };
  throw NotFound_error();
  return 0;
};

void PopList2D::SetFnx(size_t p1, size_t j, real nx)
{
  link_t &link1 = flink->At(p1)->At(j);
  size_t p2 = link1.point;
  link1.nx = nx;
  j = FindFLink(p2, p1);
  link_t &link2 = flink->At(p2)->At(j);
  link2.nx = nx;
};

void PopList2D::SetFny(size_t p1, size_t j, real ny)
{
  link_t &link1 = flink->At(p1)->At(j);
  size_t p2 = link1.point;
  link1.ny = ny;
  j = FindFLink(p2, p1);
  link_t &link2 = flink->At(p2)->At(j);
  link2.ny = ny;
};

void PopList2D::SetBnx(size_t p1, size_t j, real nx)
{
  link_t &link1 = blink->At(BPL->BIndex(p1))->At(j);
  size_t p2 = link1.point;
  link1.nx = nx;
  j = FindFLink(p2, p1);
  link_t &link2 = blink->At(BPL->BIndex(p2))->At(j);
  link2.nx = nx;
  CalcNVec(p1);
  CalcNVec(p2);
};

void PopList2D::SetBny(size_t p1, size_t j, real ny)
{
  link_t &link1 = blink->At(BPL->BIndex(p1))->At(j);
  size_t p2 = link1.point;
  link1.ny = ny;
  j = FindFLink(p2, p1);
  link_t &link2 = blink->At(BPL->BIndex(p2))->At(j);
  link2.ny = ny;
  CalcNVec(p1);
  CalcNVec(p2);
};

void PopList2D::Print(ostream &s)
{
  FORALL(i, PL->) {
    s << i << endl;
    for (size_t j = 0; j < NumFLinks(i); j++) s << FLinkPoint(i,j) << " ";
    s << endl;
    for (size_t j = 0; j < NumBLinks(i); j++) s << BLinkPoint(i,j) << " ";
    s << endl << endl;
  };
};
