// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#ifndef RoeTurkel2D_HH
#define RoeTurkel2D_HH

#include <CompressibleFFlux2D.hh>

/**
   a 2D Roe-Turkel flux for compressible gas flow.
   Thanks to Imad Elmahi and Cecile Viozat for implementing this in MOUSE.
*/
class RoeTurkel2D : public CompressibleFFlux2D {
protected:

public:
  /** */ RoeTurkel2D(ConstructArg arg);
  /** */ void operate(size_t ie, size_t ip, size_t from, size_t to, 
		      real nx, real ny, bool mod_from, bool mod_to);
};


inline void RoeTurkel2D::operate(size_t ie, size_t ip, size_t from, size_t to, 
				 real nx, real ny, bool mod_from, bool mod_to)
{
  real flux_rho=0, flux_rhou=0, flux_rhov=0, flux_rhoE=0;
  //real *flux_rhosp;
  real u_to,v_to,h_to,p_to,ep_to,rhouv_to;
  real u_from,v_from,h_from,p_from,ep_from,rhouv_from;
  //real *rhosp_to, *rhosp_from, *ysp_ave;
  real rho_ave,u_ave,v_ave,h_ave;
  real delta_rho,delta_rhou,delta_rhov,delta_rhoE;
  real hnx, hny,norme,c_speed,c_speed2;
  real eig_val1,eig_val2,eig_val3,scal,usc,u2pv2,tempo11;
  real lambda_from,lambda_to;
  real coef1,coef2,somme_coef,alpha1,alpha2,alpha3,alpha4,alpha5;
  real a1,a2,a3,a4,b1,b2,b3,b4, alpha;

  //cv for roe-turkel
  real beta2, rn, sn, tn, mustn;
  //real eig_val1_entro, eig_val3_entro;
          
  //
  //.. Compute the ROE Averages
  //
  //.... some useful quantities
  coef1 = sqrt(rho_from[ie]);
  coef2 = sqrt(rho_to[ie]);
  somme_coef = coef1 + coef2;
  u_from = rhou_from[ie]/rho_from[ie];
  v_from = rhov_from[ie]/rho_from[ie];
  h_from = (gamma * rhoE_from[ie]/rho_from[ie]) - (gamma1 / 2.0)
    * (u_from * u_from + v_from * v_from);
  u_to = rhou_to[ie]/rho_to[ie];
  v_to = rhov_to[ie]/rho_to[ie];
  h_to = (gamma * rhoE_to[ie]/rho_to[ie]) - (gamma1 / 2.0)
    * (u_to * u_to + v_to * v_to);
  
  //.... averages
  rho_ave = coef1 * coef2;
  u_ave = (coef1 * u_from + coef2 * u_to) / somme_coef;
  v_ave = (coef1 * v_from + coef2 * v_to) / somme_coef;
  h_ave = (coef1 * h_from + coef2 * h_to) / somme_coef;  

  //.. Compute Speed of sound
  scal = u_ave * nx + v_ave * ny;
  norme = sqrt(nx * nx + ny * ny);
  u2pv2 = u_ave * u_ave + v_ave * v_ave;
  c_speed = gamma1 * (h_ave - 0.5 * u2pv2);
  if(c_speed < 0.1) c_speed = 0.1;    // avoid division by 0 if critical
  c_speed = sqrt(c_speed);                    
  c_speed2 = c_speed * norme;

  //cv beta is a free parameter chosen 
  //cv on the order of the Mach number for Roe-Turkel scheme (low Mach number flow).
  //cv beta = 1. gives the Roe scheme
  //cv beta2 = beta * beta
  beta2 =  Mach_oo*Mach_oo;  
  //.. Compute the eigenvalues of the Jacobian matrix
  //cv  eig-val1 and eig_val3 must be modified 
  //cv  see RR3084 page 61 formula 116
  eig_val1 = 0.5 * ((1 + beta2) * 
		    scal - sqrt(((1 - beta2) * scal)*((1 - beta2) * scal)+
				4 * beta2 * c_speed2 * c_speed2));
  eig_val2 = scal;
  eig_val3 = 0.5 * ((1 + beta2) * 
		    scal + sqrt(((1 - beta2) * scal)*((1 - beta2) * scal)+
				4 * beta2 * c_speed2 * c_speed2));
  
  //.. Compute the ROE flux 
  //.... In this part many tests upon the eigenvalues
  //.... are done to simplify calculations                    
  //.... Here we use the two formes of the ROE flux :
  //.... phi(Wl,Wr) = F(Wl) + A-(Wroe)(Wr - Wl)
  //.... phi(Wl,Wr) = F(Wr) - A+(Wroe)(Wr - Wl)
  //
  if(eig_val2 <= 0.0) {
      p_to = gamma1 * (rhoE_to[ie] - 0.5 * (rhou_to[ie] * rhou_to[ie] + 
						  rhov_to[ie] * rhov_to[ie]) / rho_to[ie]);
      ep_to = rhoE_to[ie] + p_to;
      rhouv_to = rhou_to[ie] * v_to;
      flux_rho  = rhou_to[ie] * nx + rhov_to[ie] * ny;
      flux_rhou = (rhou_to[ie] * u_to + p_to) * nx + rhouv_to * ny;
      flux_rhov = rhouv_to * nx + (rhov_to[ie] * v_to + p_to) * ny;
      flux_rhoE = ep_to * (u_to * nx + v_to * ny);
      //
      //.... A Entropic modification
      //
      p_from = gamma1 * (rhoE_from[ie] - 0.5 * (rhou_from[ie] * rhou_from[ie] +
                 + rhov_from[ie] * rhov_from[ie]) / rho_from[ie]);        
      lambda_from = u_from * nx + v_from * ny + norme 
	* sqrt(gamma * p_from / rho_from[ie]);
      lambda_to   = u_to * nx + v_to * ny + norme
	* sqrt(gamma * p_to / rho_to[ie]);
      if ((lambda_from < 0.) && (lambda_to > 0.)) {
	eig_val3 = lambda_to * (eig_val3 - lambda_from) / (lambda_to - lambda_from);
      };
      //           
      //cv If in addition of scal <=0 we have scal + c_speed < 0
      //cv then we have three negative eigenvalues ; the flow
      //cv is supersonic ; no upwind term need to be added
      //cv
      //cv      eig_val2   eig_val3
      //cv               \   \  |t
      //cv          \     \  \  |
      //cv   eig_val1 \    \  \ |
      //cv              \   \ \ |
      //cv                \  \ \|
      //cv                  \ \\|
      //cv                    \\|
      //cv -------------------------------------------- x
      //cv
      //cv phi = F(Wl)
      //cv
      //cv If in addition of scal <=0 we have scal + c_speed > 0
      //cv then we have two negative eigenvalues and one positive
      //cv eigenvalue (eig_val3 is positive);
      //cv we need to add an upwind term.
      //cv
      //cv      eig_val2                eig_val3
      //cv               \      |t     /
      //cv          \     \     |     /
      //cv   eig_val1 \    \    |    /
      //cv              \   \   |   / 
      //cv                \  \  |  /
      //cv                  \ \ | /
      //cv                    \\|/
      //cv -------------------------------------------- x
      //cv                +
      //cv phi = F(Wl) - A delta_W
      //cv
      //cv for Roe this term can be written as
      //cv     +              -1
      //cv  - A  delta_W = - T   diag(0,0,eig_val3,0) T delta_W
      //cv  for Roe-Turkel this is replaced by
      //cv     +                 -1
      //cv  - Art delta_W = - Trt diag(0,0,eig_val3rt,0) Trt delta_W
      //cv                     -1   -1       +
      //cv                = - R   Pe  (Pe Ae)  R delta_W
      //cv  R is the jacobian matrix to go from the conservative
      //cv    to the entropic variables
      //cv  Ae is the jacobian matrix of the numerical flux in the
      //cv     entropic variables
      //cv  Pe is the Turkel preconditioner in the entropic variables
      //cv     Pe = diag(beta**2,0,0,0) where beta is on the order
      //cv     of the mach number. If beta = 1 we obtain the Roe scheme
      //cv        +                                   -1
      //cv (Pe Ae) = M diag(0,0,Max(eig_val3rt,0),0) M
      //cv                                     -1
      //cv         = M diag(0,0,eig_val3rt,0) M
      //cv  see RR3084 p 65 coeficient in front of eig_val3rt . 

      if (eig_val3 > 0.0) {
        //.. In this case A+ is obtained by multiplying the last
        //.. colomne of T-1 with the last row of T with eig_val3
	delta_rho  = rho_to[ie] - rho_from[ie];
	delta_rhou = rhou_to[ie] - rhou_from[ie];
	delta_rhov = rhov_to[ie] - rhov_from[ie];
	delta_rhoE = rhoE_to[ie] - rhoE_from[ie];
        //
	//cv in the case of the ROE scheme we have
	//cv rn = c_speed, sn = -c_speed, tn = -c_speed
	//cv see RR3084 page 61
	rn = (eig_val3 - scal * beta2) / norme;
	sn = (eig_val1 - scal * beta2) / norme;
	tn = 0.5 * (eig_val1 - eig_val3) / norme;
	mustn = - 1.0 / tn;
	scal = scal / norme;
	hnx = nx / norme;
	hny = ny / norme;
	usc = 1.0 / c_speed;
	tempo11 = gamma1 * usc;
	//.. Last columne of the matrix T-1
	a1 = mustn;
	a2 = (u_ave + rn * hnx) * mustn;
	a3 = (v_ave + rn * hny) * mustn;
	a4 = (0.5 * u2pv2  + igamma1 *
	      c_speed*c_speed + rn * scal)*mustn;
	b1 = 0.5 / beta2 * eig_val3 * 
	  (0.5 * tempo11 * u2pv2*(-sn / c_speed) - beta2 * scal);
	b2 = 0.5 / beta2 * eig_val3 * 
	  (beta2*hnx - tempo11 * u_ave * (-sn / c_speed));
	b3 = 0.5 / beta2 * eig_val3 * 
	  (beta2*hny - tempo11 * v_ave * (-sn / c_speed));
	b4 = 0.5 / beta2 * eig_val3 * tempo11 * (-sn / c_speed);
	alpha1 = b1 * delta_rho;
	alpha2 = b2 * delta_rhou;
	alpha3 = b3 * delta_rhov;
	alpha4 = b4 * delta_rhoE; 
	alpha = alpha1 + alpha2 + alpha3 + alpha4;
	alpha5 = a1 * alpha;
	//
	flux_rho  -= alpha5;
	flux_rhou -= a2 * alpha;
	flux_rhov -= a3 * alpha;                         
	flux_rhoE -= a4 * alpha;        
      };
    };
    //
  if(eig_val2 > 0.0) {
    p_from = gamma1 
      * (rhoE_from[ie] - 0.5 * (rhou_from[ie] * rhou_from[ie] +
				+ rhov_from[ie] * rhov_from[ie]) / rho_from[ie]);
    ep_from = rhoE_from[ie] + p_from;
    rhouv_from = rhou_from[ie] * v_from;
    flux_rho  = rhou_from[ie] * nx + rhov_from[ie] * ny;
    flux_rhou = (rhou_from[ie] * u_from + p_from) * nx + rhouv_from * ny;
    flux_rhov = rhouv_from * nx + (rhov_from[ie] * v_from + p_from) * ny;
    flux_rhoE = ep_from * (u_from * nx + v_from * ny);
    //
    // A Entropic modification
    //
    p_to = gamma1 
      * (rhoE_to[ie] - 0.5 * (rhou_to[ie] * rhou_to[ie] +
			      + rhov_to[ie] * rhov_to[ie]) / rho_to[ie]);        
    lambda_from = u_from * nx + v_from * ny - norme 
      * sqrt(gamma * p_from / rho_from[ie]);
    lambda_to   = u_to * nx + v_to * ny - norme
      * sqrt(gamma * p_to / rho_to[ie]);
    if ((lambda_from < 0.) && (lambda_to > 0.)) {
      //cv eig_val1_entro = lambda_from * (lambda_to - eig_val1) / (lambda_to - lambda_from);
      eig_val1 = lambda_from * (lambda_to - eig_val1) / (lambda_to - lambda_from);
    };
    //           
    //cv If in addition of scal >0 we have scal - c_speed > 0
    //cv then we have three positive eigenvalues ; the flow
    //cv is supersonic ; no upwind term need to be added
    //cv
    //cv                   eig_val1 eig_val2 eig_val3
    //cv                      |t /   /      /
    //cv                      |  /  /     /
    //cv                      | /  /    /
    //cv                      | / /   /
    //cv                      |/ /  /
    //cv                      |// /
    //cv                      |//
    //cv -------------------------------------------- x
    //cv  phi = F(Wr)
    //cv
    //cv If in addition of scal >0 we have scal - c_speed < 0
    //cv then we have two positive eigenvalues and one negative
    //cv eigenvalue (eig_val1 is negative);
    //cv we need to add an upwind term.
    //cv
    //cv            eig_val1        eig_val2 eig_val3
    //cv               \      |t     /      /
    //cv                \     |     /     /
    //cv                 \    |    /    /
    //cv                  \   |   /   /
    //cv                   \  |  /  /
    //cv                    \ | / /
    //cv                     \|//
    //cv -------------------------------------------- x
    //cv                 -
    //cv  phi = F(Wr) + A delta_W
    //cv
    //cv for Roe this term can be written as
    //cv     -              -1   
    //cv   A  delta_W =  T   diag(eig_val1,0,0,0) T delta_W
    //cv  for Roe-Turkel this is replaced by
    //cv       -               -1
    //cv    Art delta_W =   Trt diag(eig_val1rt,0,0,0) Trt delta_W
    //cv                     -1   -1       -
    //cv                =   R   Pe  (Pe Ae)  R delta_W
    //cv  R is the jacobian matrix to go from the conservative
    //cv    to the entropic variables
    //cv  Ae is the jacobian matrix of the numerical flux in the
    //cv     entropic variables
    //cv  Pe is the Turkel preconditioner in the entropic variables
    //cv     Pe = diag(beta**2,0,0,0) where beta is on the order
    //cv     of the mach number. If beta = 1 we obtain the Roe scheme
    //cv        -                                       -1
    //cv (Pe Ae) = M diag(Min(eig_val1rt,0),0,0,0),0)  M
    //cv                                            -1
    //cv         = M diag(Min(eig_val1rt,0),0,0,0) M
    //cv  see RR3084 p 65 coeficient in front of eig_val1rt . 
    if (eig_val1 < 0.0) {
      //.. In this case A+ is obtained by multiplying the first
      //.. columne of T-1 with the first row of T with eig_val1
      delta_rho  = rho_to[ie] - rho_from[ie];
      delta_rhou = rhou_to[ie] - rhou_from[ie];
      delta_rhov = rhov_to[ie] - rhov_from[ie];
      delta_rhoE = rhoE_to[ie] - rhoE_from[ie];
      //
      //cv in the case of the ROE scheme we have
      //cv rn = c_speed, sn = -c_speed, tn = -c_speed
      //cv see RR3084 page 61
      rn = (eig_val3 - scal * beta2) / norme;
      sn = (eig_val1 - scal * beta2) / norme;
      tn = 0.5 * (eig_val1 - eig_val3) / norme;
      mustn = -1.0 / tn;
      //
      scal = scal / norme;        
      hnx = nx / norme;
      hny = ny / norme;
      usc = 1.0 / c_speed;
      tempo11 = gamma1 * usc;
      //.. First colomne of the matrix T-1
      a1 = mustn;
      a2 = (u_ave + sn * hnx) * mustn;
      a3 = (v_ave + sn * hny) * mustn;
      a4 = (0.5 * u2pv2 + igamma1 * c_speed * c_speed + sn * scal) * mustn;
      //.. First row of the matrix T * eig_val1
      b1 = 0.5 / beta2 * eig_val1 * 
	(0.5 * tempo11 * u2pv2 * rn / c_speed + beta2 * scal);
      b2 = -0.5 / beta2 * eig_val1 * 
	(beta2 * hnx + tempo11 * u_ave * rn / c_speed);
      b3 = -0.5 / beta2 * eig_val1 * 
	(beta2 * hny + tempo11 * v_ave * rn / c_speed);
      b4 = 0.5 / beta2 * eig_val1 * tempo11 * rn / c_speed;
      alpha1 = b1 * delta_rho;
      alpha2 = b2 * delta_rhou;
      alpha3 = b3 * delta_rhov;
      alpha4 = b4 * delta_rhoE; 
      alpha  = alpha1 + alpha2 + alpha3 + alpha4;
      alpha5 = a1 * alpha;
      //
      flux_rho  += alpha5;
      flux_rhou += a2 * alpha;
      flux_rhov += a3 * alpha;                         
      flux_rhoE += a4 * alpha;        
    };
  };        
  //.. Add to pointwise residual
  //.... .._to-side variables
  if (mod_to) {
    res_rho[to] += flux_rho;
    res_rhou[to] += flux_rhou;
    res_rhov[to] += flux_rhov;
    res_rhoE[to] += flux_rhoE;
  };
  //.... .._from-side variables
  if (mod_from) {
    res_rho[from] -= flux_rho;
    res_rhou[from] -= flux_rhou;
    res_rhov[from] -= flux_rhov;
    res_rhoE[from] -= flux_rhoE;
  };
};

#endif









