// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include "Triangulator2D.hh"
#include <ConstructArg.hh>
#include <geometry2D.hh>

//
//.. constructor
//
Triangulator2D::Triangulator2D(ConstructArg arg) : Mesh2DAction(arg)
{
  AddClassName();
  if (LookForObject<Triangulator2D>(false, true) != NULL) {
    Error("only one 'Triangulator2D' per mesh allowed");
  };
  MESH = FindObject<Mesh2D>(false, true);
  PL = FindObject<PointList>(false, true);
  front_p1 = new TList<size_t>(200, 200);
  front_p2 = new TList<size_t>(front_p1);
};

//
//.. Act
//
void Triangulator2D::Act() {
  while (front_p1->NumActiveEntries() != 0) {
    FORALL (i_front, front_p1->) GenerateTriangle (i_front);
    front_p1->CleanUp();
  };
};

//
//.. GenerateTriangle
//
bool Triangulator2D::GenerateTriangle (size_t front) {
  int ok = 1;
  real k1, l1, k2, l2;
  real x1, y1, x2, y2, x3, y3, xj1, yj1, xj2, yj2;
  size_t pj1, pj2;
  size_t p1 = front_p1->At(front);
  size_t p2 = front_p2->At(front);
  size_t p3;
  try {
    p3 = FindBest(front);
  }
  catch (NotFound_error) {
    return false;
  };
  size_t new_front;

  // get all the coordinates
  x1 = PL->x(p1);
  y1 = PL->y(p1);
  x2 = PL->x(p2);
  y2 = PL->y(p2);
  x3 = PL->x(p3);
  y3 = PL->y(p3);

  FORALL (j_front, front_p1->) {
    k1 = -2;
    l1 = -2;
    k2 = -2;
    l2 = -2;
    pj1 = front_p1->At(j_front);
    pj2 = front_p2->At(j_front);
    //cout << pj1 << " " << pj2 << endl;
    xj1 = PL->x(pj1); 
    yj1 = PL->y(pj1); 
    xj2 = PL->x(pj2); 
    yj2 = PL->y(pj2);
    if ((pj1 != p1) && (pj1 != p3) && (pj2 != p1) && (pj2 != p3)) {
      Intersection (k1, l1, x1, y1, x3-x1, y3-y1, xj1, yj1, xj2-xj1, yj2-yj1);
    };
    if ((pj1 != p2) && (pj1 != p3) && (pj2 != p2) && (pj2 != p3)) {
      Intersection (k2, l2, x2, y2, x3-x2, y3-y2, xj1, yj1, xj2-xj1, yj2-yj1);
    };
    if ( ((k1 <= 1) && (k1 >= 0) && (l1 <= 1) && (l1 >=0)) 
	 || ((k2 <= 1) && (k2 >= 0) && (l2 <= 1) && (l2 >= 0)) ) {
      ok = 0;
      break;
    };
  };
  if (ok) {
    try {
      EliminateFront (p1, p3);
    }
    catch (NotFound_error) {
      new_front = front_p1->AddEntry();
      front_p1->At(new_front) = p1;
      front_p2->At(new_front) = p3;
    };
    try {
      EliminateFront (p3, p2);
    }
    catch (NotFound_error) {
      new_front = front_p1->AddEntry();
      front_p1->At(new_front) = p3;
      front_p2->At(new_front) = p2;
    };
    front_p1->DelEntry(front);
    // generate new triangle
    MESH->AddTri(p1, p2, p3);
  };
  return true;
};

//
//.. FindBest
//
size_t Triangulator2D::FindBest (size_t front) {
  real k, l, l_min;
  real nx, ny, xm, ym, cx, cy;
  int N = 0;
  bool found = false;
  size_t p_min;
  bool first = true;
  size_t p1 = front_p1->At(front);
  size_t p2 = front_p2->At(front);
  size_t point;

  // normal vector of frontsegment
  real fs_nx = (PL->y(p2) - PL->y(p1));
  real fs_ny = - (PL->x(p2) - PL->x(p1));
  // middle point of frontsegment
  real fs_xm = .5*(PL->x(p1) + PL->x(p2));
  real fs_ym = .5*(PL->y(p1) + PL->y(p2));
  // loop to find the best point
  FORALL (i_front, front_p1->) {
    point = front_p1->At(i_front);
    if ((point != p1) && (point != p2)) {
      N++;
      // normal vector
      nx = (PL->y(point) - PL->y(p1));
      ny = - (PL->x(point) - PL->x(p1));
      // middle point
      xm = .5*(PL->x(point) + PL->x(p1));
      ym = .5*(PL->y(point) + PL->y(p1));
      // connection vector
      cx = xm - fs_xm;
      cy = ym - fs_ym;
      real test1 = cx*fs_nx + cy*fs_ny;
      real test2 = fs_nx*fs_nx + fs_ny*fs_ny;
      if (test1 / test2 > 1e-20) {
	if (Intersection (l, k, fs_xm, fs_ym, fs_nx, fs_ny, xm, ym, nx, ny)) {
	  if (first || (l < l_min)) {
	    l_min = l;
	    p_min = point;
	    first = false;
	    found = true;
	  };
	};
      };
    };
    point = front_p2->At(i_front);
    if ((point != p1) && (point != p2)) {
      N++;
      // normal vector
      nx = (PL->y(point) - PL->y(p1));
      ny = - (PL->x(point) - PL->x(p1));
      // middle point
      xm = .5*(PL->x(point) + PL->x(p1));
      ym = .5*(PL->y(point) + PL->y(p1));
      // connection vector
      cx = xm - fs_xm;
      cy = ym - fs_ym;
      real test1 = cx*fs_nx + cy*fs_ny;
      real test2 = fs_nx*fs_nx + fs_ny*fs_ny;
      if (test1 / test2 > 1e-20) {
	if (Intersection (l, k, fs_xm, fs_ym, fs_nx, fs_ny, xm, ym, nx, ny)) {
	  if (first || (l < l_min)) {
	    l_min = l;
	    p_min = point;
	    first = false;
	    found = true;
	  };
	};
      };
    };
  };
  if (!found) {
    cout << N << endl;
    cout << "couldn't generate triangle" << endl;
    throw NotFound_error();
  };
  return p_min;
};

//
//.. EliminateFront
//
size_t Triangulator2D::EliminateFront(size_t p1, size_t p2) {
  size_t pi1, pi2;
  FORALL (i_front, front_p1->) {
    pi1 = front_p1->At(i_front);
    pi2 = front_p2->At(i_front);
    if (((pi1 == p1) && (pi2 == p2)) || ((pi1 == p2) && (pi2 == p1))) {
      front_p1->DelEntry (i_front);
      return i_front;
    };
  };
  throw NotFound_error();
};

//
//.. InitFronts
//
void Triangulator2D::InitFronts()
{
  delete front_p1;
  delete front_p2;
  front_p1 = new TList<size_t>(200, 200);
  front_p2 = new TList<size_t>(front_p1);
  PL = FindObject<PointList>(false, true);
};

//
//.. AddFront
//
void Triangulator2D::AddFront(size_t p1, size_t p2)
{
  size_t i;
  i = front_p1->AddEntry();
  front_p1->At(i) = p1;
  front_p2->At(i) = p2;
  real dx = PL->x(p2) - PL->x(p1);
  real dy = PL->y(p2) - PL->y(p1);
  MESH->SetNetDensity(p1, .5/DIST(0, 0, dx, dy));
  MESH->SetNetDensity(p2, .5/DIST(0, 0, dx, dy));
  MESH->SetMatrixComp(p1, 1, 0, 1);
  MESH->SetMatrixComp(p2, 1, 0, 1);
};







