// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#include <FindMinMax.hh>
#include <CellProjection2D.hh>

template<class POP>
AggloMesh2DBasic<POP>::AggloMesh2DBasic(ConstructArg arg) : CoarsenMesh2D<POP>(arg)
{
  AddClassName();
};

template<class POP>
void AggloMesh2DBasic<POP>::BuildBoundaries()
{
  // get the finest level agent
  MultiGridAgent<POP> *mga_f = mg_agent->Finest();

  // get this level's agent
  MultiGridAgent<POP> *mga_c;
  FindObject(mga_c, false, true);

  // a temporary array to check if a finest b.point has been added already
  TList<bool> added(mga_f->FinePL(), false);

  // a temporary array for the boundary fine -> coarse mapping
  TList<size_t> p_coarse(mga_f->FinePL());

  // a temporary array for the boundary fine <- coarse mapping (DEBUG!!)
  TList<size_t> p_fine(PL,0);

  // a loop over all finest boundary nodes
  FORALL(i_bf, mga_f->FineBPL()->) {

    // create a new point for the finest boundary node
    size_t i_f = mga_f->FineBPL()->FIndex(i_bf);
    size_t i = PL->AddEntry();
    added[i_f] = true;
    p_coarse[i_f] = i;
    p_fine[i] = i_f;
    mga_c->HideNode(i);

    // set all properties to those of the finest level
    PL->x(i) = mga_f->FinePL()->x(i_f);
    PL->y(i) = mga_f->FinePL()->y(i_f);
    PL->code(i) = mga_f->FinePL()->code(i_f);
    PL->SetIVol(i, mga_f->FinePL()->IVol(i_f));
    PL->SetChDist(i, mga_f->FinePL()->ChDist(i_f));

    // add the point to the local BoundaryPointList
    BPL->AddPoint(i);
    BPL->SetNx(i,  mga_f->FineBPL()->Nx(i_f));
    BPL->SetNy(i,  mga_f->FineBPL()->Ny(i_f));

    // compute the corresponding agglomerate
    MultiGridAgent<POP> *mga = mga_f;
    size_t agglo = i_f;
    while(mga != mga_c) {
      agglo = mga->CoarsePoint(agglo);
      mga = mga->Next();
    };
    mga_c->SetFieldNode(i, agglo); 
    // well ... hope that works !

    // add the boundary edges
    size_t num_neigh = mga_f->FinePOP()->NumBLinks(i_f);
    for (size_t j = 0; j < num_neigh; j++) {
      size_t neigh_f = mga_f->FinePOP()->BLinkPoint(i_f, j);
      if (added[neigh_f]) {
	size_t neigh = p_coarse[neigh_f];
	try {
	  POPL->FindBLink(i, neigh);
	} catch (NotFound_error) {
	  real bnx = mga_f->FinePOP()->Bnx(i_f, j);
	  real bny = mga_f->FinePOP()->Bny(i_f, j);
	  POPL->AddBLink(i, neigh, bnx, bny);
	};
      };
    };
  };

  // loop over all field nodes to compute the BRatio value
  FORALL(i, PL->) {

    // get a list of all boundary points for the field node
    TList<size_t> *b_nodes = mga_c->BoundaryPoints(i);

    real volume = 0;

    // loop over all this boundary nodes
    if (b_nodes != NULL) {
      FORALL(j, b_nodes->) {
	if (PL->code(b_nodes->At(j)) != 0) volume += PL->Vol(b_nodes->At(j));
	
	// compute the ratio and set it
	real bratio = volume/PL->Vol(i);
	mga_c->SetBRatio(i, bratio);
      };
    } else {
      mga_c->SetBRatio(i, 1);
    };
  };
};

template<class POP>
void AggloMesh2DBasic<POP>::BuildPop()
{
  // loop over all fine grid nodes
  FORALL(p1_fine, PL_F->) {

    // we will create the field links here
    // loop over all neighbours of a fine grid node
    size_t nb_neighbour = POPL_F->NumFLinks(p1_fine);
    for (size_t cur_nghb = 0; cur_nghb < nb_neighbour; cur_nghb++) {
      size_t p2_fine = POPL_F->FLinkPoint(p1_fine, cur_nghb);
      size_t p1_coarse = mg_agent->CoarsePoint(p1_fine);
      size_t p2_coarse = mg_agent->CoarsePoint(p2_fine);
      
      // if p1 and p2 belong to two different coarse nodes, contribute to that link
      if (p1_coarse != p2_coarse) {

	// only add the link if (p1_fine < p2_fine),
	// to avoid double entries!
	if (p1_fine < p2_fine) {
	  real nx, ny;
	  if (p1_coarse < p2_coarse) {
	    nx = POPL_F->Fnx(p1_fine, cur_nghb);
	    ny = POPL_F->Fny(p1_fine, cur_nghb);
	  } else {
	    nx = -POPL_F->Fnx(p1_fine, cur_nghb);
	    ny = -POPL_F->Fny(p1_fine, cur_nghb);
	  };
	  try {
	    size_t link = POPL->FindFLink(p1_coarse, p2_coarse);
	    nx += POPL->Fnx(p1_coarse, link);
	    ny += POPL->Fny(p1_coarse, link);
	    POPL->SetFnx(p1_coarse, link, nx);
	    POPL->SetFny(p1_coarse, link, ny);
	  } catch (NotFound_error) {
	    POPL->AddFLink(p1_coarse, p2_coarse, nx, ny);
	  };
	};
      };
    };
  };

  // now we have all coarse nodes
  // time to update the BoundaryPointList
  //BPL->Update();
    
  // we still have to create some boundary edges
  // every edge from the fine mesh will be found here
  // that means edges where (from == to) is true might exist

  /*  
  FORALL(cur_boundary_node, BPL_F->) {
    size_t cur_node = BPL_F->FIndex(cur_boundary_node);
    
    // loop over all neighbours of a fine grid node
    size_t nb_neighbour = POPL_F->NumBLinks(cur_node);
    for (size_t cur_nghb = 0; cur_nghb < nb_neighbour; cur_nghb++) {
      size_t neighbour = POPL_F->BLinkPoint(cur_node, cur_nghb);
      if (cur_node < neighbour) {
	cout << "add a blink " << cur_node << "->" << neighbour << endl;
	size_t p1_coarse = mg_agent->CoarsePoint(cur_node);
	size_t p2_coarse = mg_agent->CoarsePoint(neighbour);
	cout << "            " << POPL_F->Bnx(cur_node, cur_nghb) << " , "
	     << POPL_F->Bny(cur_node, cur_nghb) << endl;
	if (p1_coarse != p2_coarse) {
	  POPL->AddBLink(p1_coarse, p2_coarse, 
			 POPL_F->Bnx(cur_node, cur_nghb), 
			 POPL_F->Bny(cur_node, cur_nghb));
	} else {
	  POPL->AddBLink(p1_coarse, p2_coarse, 
			 .5*POPL_F->Bnx(cur_node, cur_nghb), 
			 .5*POPL_F->Bny(cur_node, cur_nghb));
	};
      };
    };
  };
  */
  
  // since we have all the vectors, we can now run the
  // 'Ales' way of constructing the characteristic length

  // get this level's agent
  //MultiGridAgent<POP> *mga_c = FindObject<MultiGridAgent<POP> >(false, true);
  MultiGridAgent<POP> *mga_c;
  FindObject(mga_c, false, true);

  FORALL(i,PL->) {
    // set up the projection function
    CellProjection2D fcell;
    for (size_t j = 0; j < POPL->NumFLinks(i); j++) {
      fcell.AddVector(POPL->Fnx(i,j), POPL->Fny(i,j));
    };
    TList<size_t> *bp = mga_c->BoundaryPoints(i);
    if (bp != NULL){
      FORALL(j, bp->) {
	for (size_t k = 0; k < POPL->NumBLinks(bp->At(j)); k++) {
	  fcell.AddVector(POPL->Bnx(bp->At(j),k), POPL->Bny(bp->At(j),k));
	};
      };
    };
    
    // find the max and min values
    FindMinMax<CellProjection2D> FMM(&fcell, 0, M_PI, 20, 0.05);
    real rmin = .5*FMM.Min();
    real rmax = .5*FMM.Max();
    
    // set the characteristic length to rmin
    PL->SetChDist(i, rmin);
  };
};    

template<class POP>
void AggloMesh2DBasic<POP>::ComputeCoordinates()
{
  // check if PL_F is the finest PointList
  bool is_finest = (mg_agent == mg_agent->Finest());

  // compute physical coordinates for the coarse points
  FORALL(i, PL_F->) {
    if (!mg_agent->IsHiddenNode(i)) {
      size_t ic = mg_agent->CoarsePoint(i);
      PL->x(ic) = 0;
      PL->y(ic) = 0;
    };
  };
  FORALL(i, PL_F->) {
    if (!mg_agent->IsHiddenNode(i)) {
      size_t ic = mg_agent->CoarsePoint(i);
      PL->x(ic) += PL_F->x(i) * PL_F->Vol(i);
      PL->y(ic) += PL_F->y(i) * PL_F->Vol(i);
    };
  };
  TList<bool> divided(PL, false);
  FORALL(i, PL_F->) {
    if (!mg_agent->IsHiddenNode(i)) {
      size_t ic = mg_agent->CoarsePoint(i);
      if (!divided[ic]) {
	PL->x(ic) *= PL->IVol(ic);
	PL->y(ic) *= PL->IVol(ic);
	divided[ic] = true;
      };
    };
  };
  FORALL(i, PL->) {
    TList<size_t> *bp = mg_agent->Next()->BoundaryPoints(i);
    if (bp != NULL) {
      size_t j = bp->begin();
      int code = PL->code(bp->At(j));
      int prio = BPL->Prio(bp->At(j));
      while (j < bp->end()) {
	if (BPL->Prio(bp->At(j)) > prio) {
	  prio = BPL->Prio(bp->At(j));
	  code = PL->code(bp->At(j));
	};
	j = bp->next(j);
      };
      if (bp->NumActiveEntries() == 1) {
	//PL->x(i) = PL->x(bp->At(bp->begin()));
	//PL->y(i) = PL->y(bp->At(bp->begin()));
      } else {
	real dmin1 = 1e30;
	real dmin0 = 1e30;
	real x0, y0, x1, y1;
	size_t j = bp->begin();
	while (j < bp->end()) {
	  real d = DIST(PL->x(i), PL->y(i), PL->x(bp->At(j)), PL->y(bp->At(j)));
	  if (d < dmin0) {
	    dmin1 = dmin0;
	    x1 = x0;
	    y1 = y0;
	    dmin0 = d;
	    x0 = PL->x(bp->At(j));
	    y0 = PL->y(bp->At(j));
	  } else if (d < dmin1) {
	    dmin1 = d;
	    x1 = PL->x(bp->At(j));
	    y1 = PL->y(bp->At(j));
	  };
	  j = bp->next(j);
	};
	//PL->x(i) = .5*(x0+x1);
	//PL->y(i) = .5*(y0+y1);
      };
      PL->code(i) = code;
      BPL->AddPoint(i);
      BPL->SetNx(i, 1);
      BPL->SetNy(i, 0);
    };
  };

  /*
  FORALL(i, PL->) {
    //if (PL->code(i) == 0) {
    TList<size_t> *bp = mg_agent->Next()->BoundaryPoints(i);
    if (bp == NULL) {
      PL->x(i) *= PL->IVol(i);
      PL->y(i) *= PL->IVol(i);
    } else {

      // test stuff ...!!
      PL->x(i) *= PL->IVol(i);
      PL->y(i) *= PL->IVol(i);

      size_t j = bp->begin();
      int code = PL->code(bp->At(j));
      int prio = BPL->Prio(bp->At(j));
      while (j < bp->end()) {
	if (BPL->Prio(bp->At(j)) > prio) {
	  prio = BPL->Prio(bp->At(j));
	  code = PL->code(bp->At(j));
	};
	j = bp->next(j);
      };
      PL->code(i) = code;
      int N = 0;
      real vol = 0;
      BPL->AddPoint(i);
      real nx = 0;
      real ny = 0;
      FORALL(j, bp->) {
	size_t i_b = bp->At(j);
	//PL->x(i) += PL->x(i_b)*PL->Vol(i_b);
	//PL->y(i) += PL->y(i_b)*PL->Vol(i_b);
	vol += PL->Vol(i_b);
	//here we go again!!!!!  <<<<<<<<<----------------

	for (size_t k = 0; k < POPL->NumBLinks(i_b); k++) {
	  nx += POPL->Bnx(i_b, k);
	  ny += POPL->Bny(i_b, k);
	};
	//nx += BPL->Nx(i_b);
	//ny += BPL->Ny(i_b);
      };
      //PL->x(i) /= vol;
      //PL->y(i) /= vol;
      real n_abs = DIST(0,0,nx,ny);
      BPL->SetNx(i, nx/n_abs);
      BPL->SetNy(i, ny/n_abs);
    };
  };
  */
};

template<class POP>
void AggloMesh2DBasic<POP>::Setup()
{
  // get this level's agent
  //MultiGridAgent<POP> *mga_c = FindObject<MultiGridAgent<POP> >(false, true);
  MultiGridAgent<POP> *mga_c;
  FindObject(mga_c, false, true);
  
  size_t hidden_F = 0;
  FORALL(i, PL_F->) if (mg_agent->IsHiddenNode(i)) hidden_F++;
  BuildBoundaries();
  size_t hidden_C = 0;
  FORALL(i, PL->) if (mga_c->IsHiddenNode(i)) hidden_C++;
  BuildPop();
  ComputeCoordinates();

  cout << "\nagglomerated multi-grid level created\n";
  cout << "----------------------------------------------\n";
  cout << "number of fine nodes            : " 
       << PL_F->NumActiveEntries()-hidden_F << endl;
  cout << "number of fine boundary nodes   : " 
       << BPL_F->NumActiveEntries()-hidden_F << endl;
  cout << "number of fine hidden nodes     : " 
       << hidden_F << endl;
  cout << "number of course nodes          : " 
       << PL->NumActiveEntries()-hidden_C << endl;
  cout << "number of coarse boundary nodes : " 
       << BPL->NumActiveEntries()-hidden_C << endl;
  cout << "number of coarse hidden nodes   : " 
       << hidden_C << endl;
};
