// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#if defined(FLEX2DARRAY_H)

//
//.. constructor
//
template <class T>
Flex2DArray<T>::Flex2DArray (size_t a_max_num_entries, size_t a_delta_entries)
  : List (a_max_num_entries, a_delta_entries) 
{
  count = new unsigned int [a_max_num_entries];
  value = new T* [a_max_num_entries];
  for (size_t i_entry = 0; i_entry < a_max_num_entries; i_entry++) {
    count[i_entry] = 0;
    value[i_entry] = NULL;
  };
};

template <class T>
Flex2DArray<T>::Flex2DArray (List *a_master) : List (a_master)
{
  count = new unsigned int [max_num_entries];
  value = new T* [max_num_entries];
  for (size_t i_entry = 0; i_entry < max_num_entries; i_entry++) {
    count[i_entry] = 0;
    value[i_entry] = NULL;
  };
};

//
//.. destructor
//
template <class T>
Flex2DArray<T>::~Flex2DArray () {
  for (size_t i = 0; i < max_num_entries; i++) {
    if (value[i] != NULL) delete [] value[i];
  };
  //delete [] value;
  delete [] count;
};

//
//.. NewLast
//
//.... dynamic extension of first dimension might move in this method !!!@@
//
template <class T>
T& Flex2DArray<T>::NewLast (size_t i) {
  T *new_value;
  new_value = new T [count[i] + 1];
  for (unsigned int j = 0; j < count[i]; j++) new_value[j] = value[i][j];
  delete [] value[i];
  value[i] = new_value;
  count[i]++;
  return new_value[count[i] - 1];
};

//
//.. NewAt
//
template <class T>
T& Flex2DArray<T>::NewAt (size_t i, unsigned int at_j) {
  T *new_value;
  unsigned int j;
  new_value = new T [count[i] + 1];
  for (j = 0; j < at_j; j++) new_value[j] = value[i][j];
  for (j = at_j; j < count[i]; j++) new_value[j + 1] = value[i][j];
  delete [] value[i];
  value[i] = new_value;
  count[i]++;
  return new_value[at_j];
};

//
//.. Resize
//
template <class T>
void Flex2DArray<T>::Resize (size_t i, unsigned int size) {
  delete [] value[i];
  value[i] = new T [size];
  count[i] = size;
};

//
//.. DelItem
//
// no error checking if (i, at_j) is illegal !!@@??
//
template <class T>
void Flex2DArray<T>::DelItem (size_t i, unsigned int at_j) {
  T *new_value;
  unsigned int j;
  if (count[i] > 1) {
    new_value = new T [count[i] - 1];
    for (j = 0; j < at_j; j++) new_value[j] = value[i][j];
    for (j = at_j; j < count[i] - 1; j++) new_value[j] = value[i][j + 1];
  } else {
    new_value = NULL;
  };
  delete [] value[i];
  value[i] = new_value;
  count[i]--;
};

//
//.. DelSubList
//
template <class T>
void Flex2DArray<T>::DelSubList(size_t i)
{
  delete [] value[i];
  value[i] = NULL;
  count[i] = 0;
};


//
//.. FindItem
//
template <class T>
size_t Flex2DArray<T>::FindItem (size_t entry, T item) const
{
  for (unsigned int i = 0; i < count[entry]; i++) if (value[entry][i] == item) return i;
  throw NotFound_error();
};

//
//.. Extend
//.... virtual method from class "List"
//
template <class T>
void Flex2DArray<T>::Extend (size_t delta) {
  size_t size = max_num_entries + delta;
  //printf ("inside TList::Extend\n");
  T **new_value;
  unsigned int *new_count;
  new_value = new T* [size];
  new_count = new unsigned int [size];
  for (size_t i_entry = 0; i_entry < num_entries + delta; i_entry++) {
    if (i_entry < num_entries) {
      new_value[i_entry] = value[i_entry];
      new_count[i_entry] = count[i_entry];
    } else {
      new_value[i_entry] = NULL;
      new_count[i_entry] = 0;
    };
  };
  //max_num_entries += delta;
  delete [] value;
  delete [] count;
  value = new_value;
  count = new_count;
};

//
//.. InitEntry
//
template <class T>
void Flex2DArray<T>::InitEntry (size_t entry) {
  value[entry] = NULL;
  count[entry] = 0;
};

//
//.. CopyEntry
//
template <class T>
void Flex2DArray<T>::CopyEntry (size_t source_point, size_t dest_point) {
  unsigned int i;
  delete [] value[dest_point];
  value[dest_point] = new T [count[source_point]];
  count[dest_point] = count[source_point];
  for (i = 0; i < count[source_point]; i++) {
    value[dest_point][i] = value[source_point][i];
  };
};


#endif







