// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#if !defined(FLEX2DARRAY_H)
#define FLEX2DARRAY_H

template <class T> class Flex2DArray;

#include <List.hh>

/**
   a dynamic two-dimensional array.
   The second dimension of this array may vary over the first dimension. That
   means every entry in the first dimension represents a one-dimensional array
   of variable length. 
*/
template <class T>
class Flex2DArray : public List 
{
protected:
  /** array with the number of entries for the second dimension */
  unsigned int *count;

  /** array with the stored values */
  T **value;

public:
  /** constructor.
    @param a_max_num_entries the initial maximal number of entries
    @param a_delta_entries the initial increment */
  Flex2DArray (size_t a_max_num_entries, size_t a_delta_entries);

  /** constructor.
    @param a_master master List for dynamics */
  Flex2DArray (List *a_master);
  Flex2DArray () : List () {};

  /** destructor. */
  ~Flex2DArray ();

  /** Add a new entry.
    This method adds a new entry at the end of a subarray.
    @param i the index in the first dimension
    @return a reference to the new entry */
  T& NewLast (size_t i);

  /** Add a new entry.
    This method adds a new entry at a given position in a subarray.
    @param i1 the index in the first dimension
    @param i2 the index in the second dimension */
  T& NewAt (size_t i1, unsigned int i2);

  /** Resize a sub-array.
    This method deletes a subarray and creates a new one with a given length.
    @param i the index in the first dimension
    @param l the length of the new subarray */
  void Resize (size_t i, unsigned int l);

  /** The Size of a subarray.
    @param i the index in the first dimension
    @return the number of entries in the subarray */
  size_t Size (size_t i) const { return count[i]; };

  /** Access to an entry.
    This method provides access to a specific entry of a subarray.
    @param i1 the index in the first dimension
    @param i2 the index in the second dimension 
    @return a reference to the entry */
  T& At (size_t i1, unsigned int i2) { return value[i1][i2]; };

  /** Access to an entry.
    This method provides access to the last entry of a subarray.
    @param i the index in the first dimension
    @return a reference to the entry */
  T& Last (size_t i) { return value[i][count[i] - 1]; };

  /** Delete an entry.
    @param i1 the index in the first dimension
    @param i2 the index in the second dimension */
  void DelItem (size_t i1, unsigned int i2);

  /** delete a sublist. */
  void DelSubList(size_t i); 

  /** Find an entry.
    @param i the index in the first dimension
    @param v the value to search for
    @return the index in the second dimension of an entry equal to v
  */
  size_t FindItem (size_t i, T v) const;

  // virtual methods from class "List"

  /** */ virtual void Extend (size_t);
  /** */ virtual void CopyEntry (size_t, size_t);
  /** */ virtual void InitEntry (size_t);
};

#include "Flex2DArray.cc"

#endif







