#include "MultiGridAgent.hh"
#include <CoarsenMesh2D.hh>

template<class POP>
MultiGridAgent<POP>::MultiGridAgent(carg_t arg) : MObject(arg)
{
  AddClassName();
  PL_fine = FindObject<PointList>(false, true);
  BPL_fine = FindObject<BoundaryPointList>(false, true);
  POPL_fine = FindObject<POP>(false, true);
  VL_fine = FindObject<VarList>(false, true);
  coarse_point = new TList<size_t>(PL_fine);
  hidden = new TList<bool>(PL_fine, false);
  VL_coarse = NULL;
  PL_coarse = NULL;
  BPL_coarse = NULL;
  POPL_coarse = NULL;
  MGA_coarse = NULL;

  CoarsenMesh2D<POP> *cmesh = LookForObject<CoarsenMesh2D<POP> >(false, true);
  if (cmesh != NULL) {
    MGA_fine = cmesh->FineMGA();
    MGA_fine->MGA_coarse = this;
    field_point = new TList<size_t>(BPL_fine);
    boundary_points = new TList<TList<size_t>*>(PL_fine, NULL);
    bratio = new TList<real>(PL_fine, 1.0);
  } else {
    MGA_fine = NULL;
    field_point = NULL;
    boundary_points = NULL;
    bratio = NULL;
  };
};

template<class POP>
VarList* MultiGridAgent<POP>::CoarseVL()
{
  if (VL_coarse == NULL) {
    string err_text = "The coarse VarList pointer is not available yet.\n";
    err_text += "It has not been set by the coarse mesh, so try to access it later.";
    Error(err_text);
  };
  return VL_coarse;
};

template<class POP>
PointList* MultiGridAgent<POP>::CoarsePL()
{
  if (PL_coarse == NULL) {
    string err_text = "The coarse PointList pointer is not available yet.\n";
    err_text += "It has not been set by the coarse mesh, so try to access it later.";
    Error(err_text);
  };
  return PL_coarse;
};

template<class POP>
POP* MultiGridAgent<POP>::CoarsePOP()
{
  if (POPL_coarse == NULL) {
    string err_text = "The coarse POP-list pointer is not available yet.\n";
    err_text += "It has not been set by the coarse mesh, so try to access it later.";
    Error(err_text);
  };
  return POPL_coarse;
};

template<class POP>
BoundaryPointList* MultiGridAgent<POP>::CoarseBPL()
{
  if (BPL_coarse == NULL) {
    string err_text = "The coarse BoundaryPointList pointer is not available yet.\n";
    err_text += "It has not been set by the coarse mesh, so try to access it later.";
    Error(err_text);
  };
  return BPL_coarse;
};

template<class POP>
MultiGridAgent<POP>* MultiGridAgent<POP>::Finest()
{
  MultiGridAgent<POP> *mga = this;
  while (mga->MGA_fine != NULL) mga = mga->MGA_fine;
  return mga;
};

template<class POP>
MultiGridAgent<POP>* MultiGridAgent<POP>::Coarsest()
{
  MultiGridAgent<POP> *mga = this;
  while (mga->MGA_coarse != NULL) mga = mga->MGA_coarse;
  return mga;
};

template<class POP>
int MultiGridAgent<POP>::Level()
{
  int level = 0;
  MultiGridAgent<POP> *mga = Finest();
  while (mga != this) {
    mga = mga->Next();
    level++;
  };
  return level;
};

template<class POP>
void MultiGridAgent<POP>::HandleEvent(Event &E)
{
  MObject::HandleEvent(E);
  if (!E.Done()) {
    if (E.name == "CoarsenMesh2D_created") {
      if (MGA_fine != NULL) {
	Error("there is only one CoarsenMesh2D allowed per grid level");
      };
      CoarsenMesh2D<POP> *cmesh = FindObject<CoarsenMesh2D<POP> >(false, true);
      MGA_fine = cmesh->FineMGA();
      MGA_fine->MGA_coarse = this;
      field_point = new TList<size_t>(BPL_fine);
      boundary_points = new TList<TList<size_t>*>(PL_fine, NULL);
      bratio = new TList<real>(PL_fine, 1.0);
    };
  };
};

template<class POP> 
void MultiGridAgent<POP>::SetFieldNode(size_t i_b, size_t i_f)
{
  field_point->At(BPL_fine->BIndex(i_b)) = i_f;
  if (boundary_points->At(i_f) == NULL) {
    boundary_points->At(i_f) = new TList<size_t>(1,1);
  };
  boundary_points->At(i_f)->NewEntry() = i_b;
};

template<class POP>
void MultiGridAgent<POP>::Write(string name)
{
  char word[200];
  strcpy(word, "");
  sprintf(word, "%s.L%u_agglo", name.c_str(), Level());
  ofstream f(word);
  size_t N = 0;
  FORALL(i, PL_fine->) if (!IsHiddenNode(i)) N++;
  f << N << endl;
  FORALL(i, PL_fine->) {
    if (!IsHiddenNode(i)) f << i << " " << CoarsePoint(i) << endl;
  };
};

template<class POP>
MultiGridAgent<POP>* MultiGridAgent<POP>::FindAgent(MObject *obj, int level)
{
  MultiGridAgent<POP> *mga;
  obj->FindObject(mga, false, true);
  mga = mga->Finest();
  for (int i = 0; i < level; i++) {
    if (mga == NULL) obj->Error("requesting a non existent level");
    mga = mga->Next();
  };
  return mga;
};
