#ifndef MultiGridAgent_HH
#define MultiGridAgent_HH

template<class POP> class MultiGridAgent;

#include <MObject.hh>
#include <TList.hh>
#include <VarList.hh>
#include <BoundaryPointList.hh>

/**
   A MultiGridAgent is responsible to provide the necessary information
   to go from a fine to a coarse grid. It has to be declared in the fine
   grid and will then be passed to the coarse grid. However it can only be
   used after both grids made the necessary setting (e.g. connections from
   fine to coarse nodes).
*/
template<class POP>
class MultiGridAgent : public MObject
{
protected:
  /** the fine grid PointList */ 
  PointList *PL_fine;

  /** the fine grid BoundaryPointList */ 
  BoundaryPointList *BPL_fine;

  /** the fine grid POP-list */
  POP *POPL_fine;

  /** the fine grid VarList */
  VarList *VL_fine;
  
  /** the coarse grid PointList */ 
  PointList *PL_coarse;

  /** the coarse grid BoundaryPointList */ 
  BoundaryPointList *BPL_coarse;

  /** the coarse grid POP-list */
  POP *POPL_coarse;

  /** the coarse grid VarList */
  VarList *VL_coarse;
  
  /** the corresponding coarse nodes of the fine nodes */ 
  TList<size_t> *coarse_point;

  /** the corresponding field nodes of the boundary nodes */
  TList<size_t> *field_point;

  /** the ratio of 'boundary volume' and 'field volume' for every field node */
  TList<real> *bratio;
  
  /** put something in here! */ 
  TList<bool> *hidden;

  /** the corresponding boundary nodes of the field nodes */
  TList<TList<size_t>*> *boundary_points;

  /** the next finer agent, if this is not the finest (NULL in this case) */
  MultiGridAgent *MGA_fine;

  /** the next coarser agent, if this is not the coarsest (NULL in this case) */
  MultiGridAgent *MGA_coarse;

public:
  /** */ MultiGridAgent(carg_t arg);

  /**
     get the corresponding coarse node.
     @param i the index of a fine node
     @return it's corresponding coarse node
  */ 
  size_t CoarsePoint(size_t i) { 
#ifdef MDEBUG
    if ((this != Finest()) && (PL_fine->code(i) != 0)) {
      Error("cannot get a coarse node for a 'hidden' boundary node");
    };
#endif
    return coarse_point->At(i); 
  };

  /**
     get the corresponding field node of a boundary node.
     @param i_b the boundary node
     @return the field node
  */
  size_t FieldPoint(size_t i_b) { 
#ifdef MDEBUG
    if (this == Finest()) Error("FieldPoint is not useful on the finest grid");
    if (field_point == NULL) Error("field_point has not been set yet");
#endif
    return field_point->At(BPL_fine->BIndex(i_b)); 
  };

  /**
     get the ratio between field and boundary volumes for a field node
     @param i the field node
     @return the ratio
  */
  real BRatio(size_t i) { return bratio->At(i); };

  /**
     set the ratio between field and boundary volumes for a field node
     @param i the field node
     @param br the ration
  */
  void SetBRatio(size_t i, real br) { bratio->At(i) = br; };

  /**
     get a list of the corresponding boundary nodes of an agglomerate cell.
     @param i_f the field node
     @return a pointer to a list of boundary nodes
  */
  TList<size_t>* BoundaryPoints(size_t i_f) { return boundary_points->At(i_f); };

  /**
     check if a point is a 'hidden' boundary node
     @param the index of the point
     @return true if it is a hidden boundary node
  */
  bool IsHiddenNode(size_t i) { return hidden->At(i); };
  /*
    if (PL_fine->code(i) == 0) return false;
    if (this == Finest()) return false;
    if (boundary_points->At(i) != NULL) return false;
    return true;
  };
  */

  /** */
  void HideNode(size_t i) { hidden->At(i) = true; };

  /**
     set the corresponding coarse node.
     @param i_fine the index of a fine node
     @return i_coarse corresponding coarse node
  */ 
  void SetCoarsePoint(size_t i_fine, size_t i_coarse) { 
    coarse_point->At(i_fine) = i_coarse;
  };

  /**
     set the corresponding field node of a boundary node.
     This method will also update the field -> boundary relation.
     @param i_b the boundary node
     @param i_f the field node
  */
  void SetFieldNode(size_t i_b, size_t i_f);

  /**
     get a pointer to the fine grid's PointList
  */
  PointList* FinePL() { return PL_fine; };

  /**
     get a pointer to the fine grid's POP-list
  */
  POP* FinePOP() { return POPL_fine; };

  /**
     get a pointer to the fine grid's VarList
  */
  VarList* FineVL() { return VL_fine; };

  /**
     get a pointer to the fine grid's BoundaryPointList
  */
  BoundaryPointList* FineBPL() { return BPL_fine; };

  /**
     get a pointer to the coarse grid's VarList
  */
  VarList* CoarseVL();

  /**
     get a pointer to the coarse grid's PointList
  */
  PointList* CoarsePL();

  /**
     get a pointer to the coarse grid's BoundaryPointList
  */
  BoundaryPointList* CoarseBPL();

  /**
     get a pointer to the coarse grid's POP-List
  */
  POP* CoarsePOP();

  /**
     set the coarse grid PointList pointer
  */
  void SetCoarsePL(PointList *PL) { PL_coarse = PL; };

  /**
     set the coarse grid BoundaryPointList pointer
  */
  void SetCoarseBPL(BoundaryPointList *BPL) {
    BPL_coarse = BPL;
  };

  /**
     set the coarse grid POP-list pointer
  */
  void SetCoarsePOP(POP *POPL) { POPL_coarse = POPL; };

  /**
     set the coarse grid VarList pointer
  */
  void SetCoarseVL(VarList *VL) { 
    VL_coarse = VL; 
    VL_coarse->AddField("mg_source");
  };

  /**
     get the finest MultiGridAgent
  */
  MultiGridAgent<POP>* Finest();

  /**
     get the next level's MultiGridAgent
  */
  MultiGridAgent<POP>* Next() { return MGA_coarse; };

  /**
     get the previous level's MultiGridAgent
  */
  MultiGridAgent<POP>* Prev() { return MGA_fine; };

  /**
     get the coarset MultiGridAgent
  */
  MultiGridAgent<POP>* Coarsest();

  /**
     get the grid level
     @return the grid level this agent is in
  */
  int Level();

  /**
     write the fine -> coarse information to a file
     @param name the base name of the file
  */
  void Write(string name);

  /**
     the event handler.
     will recact on:\\
     "CoarsenMesh2D_created"\\
  */
  virtual void HandleEvent(Event &E);

  /**
     Find a particular MultiGridAgent.
     @param obj an object in the same mesh as the agent we are looking for
     @param level the level of the agent
     @return a pointer to the agent
  */
  static MultiGridAgent<POP>* FindAgent(MObject *obj, int level);
};

#include "MultiGridAgent.cc"

#endif
