// !!
// This is a part of MOUSE, a library for PDE's on unstructured grids
// Copyright (C) 1999 Oliver Gloth <oliver@vug.uni-duisburg.de>
// Institut fuer Verbrennung und Gasdynamik (Universitaet Duisburg)
// institute for combustion and gas dynamics (university of Duisburg)
// Thursday, 1 April, 1999 Duisburg, Germany
//
// please see http://www.vug.uni-duisburg.de/MOUSE for more information
// please send any questions or suggestions to mouse@www.vug.uni-duisburg.de
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// !!

#if !defined(TLIST_H)
#define TLIST_H

#include <cstddef>

template <class T> class TList;
template <class T, size_t MNE, size_t DE> class TTList;

#include "List.hh"

/**
   A self-allocating one-dimensional array.
*/
template <class T>
class TList : public List 
{
protected:
  /** array of the stored values */
  T *value;
  
  /** extends the array value.
      @param delta the increment */ 
  virtual void Extend (size_t delta);

  /** initialize an entry with the default value 
      @param i the entry to initialize */ 
  virtual void InitEntry (size_t i) {
    value[i] = default_value;
  };
  
public:
  /** the default value for new entries */
  T default_value;
  
  /** @param a_max_num_entries the initial size of the value field
      @param a_delta_entries the increment for expansion */
  TList (size_t a_max_num_entries, size_t a_delta_entries);
  
  /** @param a_max_num_entries the initial maximal number of entries
      @param a_delta_entries the increment for expansion
      @param a_default_value for new entries */
  TList (size_t a_max_num_entries, size_t a_delta_entries, T a_default_value);
  
  /** @param a_master a master for this List. */
  TList (List *a_master);
  
  /** @param a_master a master for this List.
      @param a_default_value the initial value for new entries */
  TList (List *a_master, T a_default_value);
  
  /** */ TList();

  /** */ ~TList ();
  
  /** Initialize all active entries with a value.
    @param v the value for the initilaziation */
  void InitAllEntries (T v);

  /** Allocate a new entry.
    @return the index of the new entry */
  T& NewEntry () { 
    size_t i = AddEntry ();  
    return value[i];  
  };

  /** The Value of an entry.
    @param i the index of the entry
    @return a reference to the value of this entry */
  T& At (size_t i) { return value[i]; };

  /** The Value of an entry.
    @param i the index of the entry
    @return the value of this entry */
  T At (size_t i) const { 
#ifdef MDEBUG
    if ((i < 0) || (i >= num_entries)) {
      cerr << "index " << i << "out of bounds" << endl;
      throw InvalidIndex_error(i);
    };
    if (!IsActive(i)) {
      cerr << "the entry number " << i << "is inactive" << endl;
      throw InvalidIndex_error(i);
    };
#endif
    return value[i]; 
  };

  /** The Value of an entry.
      @param i the index of the entry
      @return a reference to the value of this entry */
  T& operator[] (size_t i) { return value[i]; };

  /** The Value of an entry.
      @param i the index of the entry
      @return the value of this entry */
  const T& operator[] (size_t i) const { return value[i]; };

  /** Find an entry.
    @param v the value to search for
    @return the index of the entry */
  size_t FindItem (T v) const;

  /** Check if an entry is in the list */
  bool Has(T v) const;

  /** Get a pointer to the data field */
  T* ValuePointer() { return value; };

  /** */ virtual void CopyEntry (size_t source_point, size_t dest_point) {
    List::CopyEntry(source_point, dest_point);
    value[dest_point] = value[source_point];
  };

  /**
     exchange two entries
     @param i1 the first entry
     @param i2 the second entry
   */ 
  void Swap(size_t i1, size_t i2) {
    T h = At(i1);
    At(i1) = At(i2);
    At(i2) = h;
  };

  /** */ void operator=(const TList<T> &other);
};


/**(tlists)\\
 */
template<class T, size_t MNE, size_t DE>
class TTList : public TList<T>
{
public:
  /** */ TTList() : TList<T>(MNE, DE) {};
};

#include "TList.cc"

#endif







